import math
import os
import warnings

import matplotlib
import numpy as np
from functools import cached_property
from echo import delay_callback
from astropy.visualization import ManualInterval, ContrastBiasStretch
from glue.core.subset_group import GroupedSubset
from glue.config import stretches as glue_stretches
from glue.viewers.histogram.state import HistogramViewerState
from glue.viewers.histogram.state import HistogramLayerState as BqplotHistogramLayerState
from glue.viewers.scatter.state import ScatterViewerState
from glue.viewers.profile.state import ProfileViewerState, ProfileLayerState
from glue.viewers.image.state import ImageSubsetLayerState, ImageViewerState
from glue.viewers.scatter.state import ScatterLayerState as BqplotScatterLayerState
from glue.viewers.image.composite_array import COLOR_CONVERTER
from glue_jupyter.bqplot.image.state import BqplotImageLayerState
from glue_jupyter.common.toolbar_vuetify import read_icon
from scipy.interpolate import PchipInterpolator
from traitlets import Any, Dict, Float, Bool, Int, List, Unicode, observe

from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, ViewerSelectMixin, LayerSelect,
                                        PlotOptionsSyncState, Plot,
                                        skip_if_no_updates_since_last_active, with_spinner)
from jdaviz.core.events import ChangeRefDataMessage, ViewerAddedMessage
from jdaviz.core.user_api import PluginUserApi
from jdaviz.core.tools import ICON_DIR
from jdaviz.core.custom_traitlets import IntHandleEmpty
from jdaviz.core.sonified_layers import SonifiedLayerState
# by importing from utils, glue_colormaps will include the custom Random colormap
from jdaviz.utils import is_not_wcs_only, cmap_samples, glue_colormaps

__all__ = ['PlotOptions']

RANDOM_SUBSET_SIZE = 10_000


class SplineStretch:
    """
    A class to represent spline stretches.

    Attributes
    ----------
    k : int
        Degree of the smoothing spline. Default is 3.
    bc_type : str or None
        Boundary condition type. Default is None.
    t : array-like or None
        Array of knot positions. Default is None.
    x : array-like
        The x-coordinates of the data points.
    y : array-like
        The y-coordinates of the data points.
    spline : object
        Interpolating spline.

    Raises
    ------
    ValueError
        If `x` and `y` have different lengths.
    """

    def __init__(self):
        # Default x, y values(0-1) range chosen for a typical initial spline shape.
        # Can be modified if required.
        self._x = np.array([0, 0.1, 0.2, 0.7, 1])
        self._y = np.array([0, 0.05, 0.3, 0.9, 1])
        self.update_knots(self._x, self._y)

    @property
    def knots(self):
        return (self._x, self._y)

    @knots.setter
    def knots(self, value):
        x, y = value
        if len(x) != len(y):
            # Silently return
            return
        self.update_knots(x, y)

    def __call__(self, values, out=None, clip=False):
        # For our uses, we can ignore `out` and `clip`, but those would need
        # to be implemented before contributing this class upstream.
        return self.spline(values)

    def update_knots(self, x, y):
        self._x = x
        self._y = y
        self.spline = PchipInterpolator(self._x, self._y)


# Add the spline stretch to the glue stretch registry if not registered
if "spline" not in glue_stretches:
    glue_stretches.add("spline", SplineStretch, display="Spline")


def _round_step(step):
    # round the step for a float input
    if step <= 0:
        return 1e-6, 6
    decimals = -int(np.log10(abs(step))) + 1 if step != 0 else 6
    if decimals < 0:
        decimals = 0
    return float(np.round(step, decimals)), decimals


@tray_registry('g-plot-options', label="Plot Options",
               category='core', sidebar='settings', subtab=0)
class PlotOptions(PluginTemplateMixin, ViewerSelectMixin):
    """
    The Plot Options Plugin gives access to per-viewer and per-layer options and enables
    setting across multiple viewers/layers simultaneously.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``viewer`` (:class:`~jdaviz.core.template_mixin.ViewerSelect`):
    * ``viewer_multiselect``
    * ``layer`` (:class:`~jdaviz.core.template_mixin.LayerSelect`):
    * ``layer_multiselect``
    * :meth:`select_all`
    * ``subset_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      whether a subset should be visible.
    * ``subset_color`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``subset_opacity`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``axes_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``line_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``line_color`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``line_width`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``line_opacity`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``line_as_steps`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``uncertainty_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Imviz
    * ``stretch_function`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``stretch_preset`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``stretch_vmin`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``stretch_vmax`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``stretch_hist_zoom_limits`` : whether to show the histogram for the current zoom
      limits instead of all data within the layer; not exposed for Specviz.
    * ``stretch_hist_nbins`` : number of bins to use in creating the histogram; not exposed
      for Specviz.
    * ``stretch_curve_visible`` : bool
      whether the stretch histogram's colormap "curve" is visible; not exposed for Specviz.
    * ``image_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      whether the image bitmap is visible; not exposed for Specviz.
    * ``image_color_mode`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``image_color`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz.  This only applies when ``image_color_mode`` is "Color".
    * ``image_colormap`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. This only applies when ``image_color_mode`` is "Colormap".
    * ``image_opacity`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. Valid values are between 0 and 1, inclusive. Default is 1.
    * ``image_contrast`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. Valid values are between 0 and 4, inclusive. Default is 1.
    * ``image_bias`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. Valid values are between 0 and 1, inclusive. Default is 0.5.
    * ``contour_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      whether the contour is visible; not exposed for Specviz
    * ``contour_mode`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz
    * ``contour_min`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. This only applies when ``contour_mode`` is "Linear".
    * ``contour_max`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. This only applies when ``contour_mode`` is "Linear".
    * ``contour_nlevels`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. This only applies when ``contour_mode`` is "Linear".
    * ``contour_custom_levels`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. This only applies when ``contour_mode`` is "Custom".
    * ``volume_level`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. Set the volume for the selected sonified layer.
    * ``sonified_audible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      not exposed for Specviz. Set if the selected sonified layer will output audio.
    * ``xatt`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Controls the x-axis attribute for scatter
      and histogram plots.
    * ``yatt`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Controls the y-axis attribute for scatter plots.
    * ``hist_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Whether histogram is visible.
    * ``hist_color`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Controls histogram color.
    * ``hist_opacity`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Controls histogram opacity.
    * ``hist_xlog`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Controls x-axis log scaling for histograms.
    * ``hist_ylog`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Controls y-axis log scaling for histograms.
    * ``hist_n_bin`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Controls number of bins for histograms.
    * ``hist_x_min`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Controls x-axis minimum value for histograms.
    * ``hist_x_max`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Controls x-axis maximum value for histograms.
    * ``hist_cumulative`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Controls cumulative histogram display.
    * ``hist_normalize`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Controls histogram normalization.
    * ``hist_update_bins_on_reset_limits``
      (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for deconfigged. Controls automatic bin updates when limits are reset.
    * ``marker_visible`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for Imviz and deconfigged. Whether markers are visible.
    * ``marker_fill`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for Imviz and deconfigged. Controls marker fill.
    * ``marker_opacity`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for Imviz and deconfigged. Controls marker opacity.
    * ``marker_size`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for Imviz and deconfigged. Controls marker size.
    * ``marker_size_scale`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for Imviz and deconfigged. Controls marker size scaling.
    * ``marker_color_mode`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for Imviz and deconfigged. Controls marker color mode.
    * ``marker_color`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for Imviz and deconfigged. Controls marker color.
    * ``marker_color_col`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for Imviz and deconfigged. Controls marker color column.
    * ``marker_colormap`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for Imviz and deconfigged. Controls marker colormap.
    * ``marker_colormap_vmin`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for Imviz and deconfigged. Controls marker colormap minimum value.
    * ``marker_colormap_vmax`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
      only exposed for Imviz and deconfigged. Controls marker colormap maximum value.
    """
    template_file = __file__, "plot_options.vue"
    uses_active_status = Bool(True).tag(sync=True)

    # read-only display units
    display_units = Dict().tag(sync=True)

    viewer_limits = Dict().tag(sync=True)

    layer_multiselect = Bool(False).tag(sync=True)
    layer_items = List().tag(sync=True)
    layer_selected = Any().tag(sync=True)  # Any needed for multiselect

    xatt_value = Unicode().tag(sync=True)
    xatt_sync = Dict().tag(sync=True)

    yatt_value = Unicode().tag(sync=True)
    yatt_sync = Dict().tag(sync=True)

    # profile/line viewer/layer options:
    line_visible_value = Bool().tag(sync=True)
    line_visible_sync = Dict().tag(sync=True)

    line_color_value = Any().tag(sync=True)
    line_color_sync = Dict().tag(sync=True)

    line_width_value = Int().tag(sync=True)
    line_width_sync = Dict().tag(sync=True)

    line_opacity_value = Float().tag(sync=True)
    line_opacity_sync = Dict().tag(sync=True)

    line_as_steps_value = Bool().tag(sync=True)
    line_as_steps_sync = Dict().tag(sync=True)

    uncertainty_visible_value = Int().tag(sync=True)
    uncertainty_visible_sync = Dict().tag(sync=True)

    x_min_value = Float().tag(sync=True)
    x_min_sync = Dict().tag(sync=True)

    x_max_value = Float().tag(sync=True)
    x_max_sync = Dict().tag(sync=True)

    y_min_value = Float().tag(sync=True)
    y_min_sync = Dict().tag(sync=True)

    y_max_value = Float().tag(sync=True)
    y_max_sync = Dict().tag(sync=True)

    x_bound_step = Float(0.1).tag(sync=True)  # dynamic based on maximum value
    y_bound_step = Float(0.1).tag(sync=True)  # dynamic based on maximum value

    zoom_center_x_value = Float().tag(sync=True)
    zoom_center_x_sync = Dict().tag(sync=True)

    zoom_center_y_value = Float().tag(sync=True)
    zoom_center_y_sync = Dict().tag(sync=True)

    zoom_radius_value = Float().tag(sync=True)
    zoom_radius_sync = Dict().tag(sync=True)

    zoom_step = Float(1).tag(sync=True)

    # scatter/marker options
    marker_visible_value = Bool().tag(sync=True)
    marker_visible_sync = Dict().tag(sync=True)

    marker_fill_value = Bool().tag(sync=True)
    marker_fill_sync = Dict().tag(sync=True)

    marker_opacity_value = Float().tag(sync=True)
    marker_opacity_sync = Dict().tag(sync=True)

    marker_size_mode_value = Unicode().tag(sync=True)
    marker_size_mode_sync = Dict().tag(sync=True)

    marker_size_value = Float().tag(sync=True)
    marker_size_sync = Dict().tag(sync=True)

    marker_size_scale_value = Float().tag(sync=True)
    marker_size_scale_sync = Dict().tag(sync=True)

    marker_size_col_value = Unicode().tag(sync=True)
    marker_size_col_sync = Dict().tag(sync=True)

    marker_size_vmin_value = Float().tag(sync=True)
    marker_size_vmin_sync = Dict().tag(sync=True)

    marker_size_vmax_value = Float().tag(sync=True)
    marker_size_vmax_sync = Dict().tag(sync=True)

    marker_color_mode_value = Unicode().tag(sync=True)
    marker_color_mode_sync = Dict().tag(sync=True)

    marker_color_value = Any().tag(sync=True)
    marker_color_sync = Dict().tag(sync=True)

    marker_color_col_value = Unicode().tag(sync=True)
    marker_color_col_sync = Dict().tag(sync=True)

    marker_colormap_value = Unicode().tag(sync=True)
    marker_colormap_sync = Dict().tag(sync=True)

    marker_colormap_vmin_value = Float().tag(sync=True)
    marker_colormap_vmin_sync = Dict().tag(sync=True)

    marker_colormap_vmax_value = Float().tag(sync=True)
    marker_colormap_vmax_sync = Dict().tag(sync=True)

    # image viewer/layer options
    active_layer = Unicode().tag(sync=True)
    stretch_function_value = Unicode().tag(sync=True)
    stretch_function_sync = Dict().tag(sync=True)

    stretch_preset_value = Any().tag(sync=True)  # glue will pass either a float or string
    stretch_preset_sync = Dict().tag(sync=True)

    stretch_vstep = Float(0.1).tag(sync=True)  # dynamic based on full range from image

    stretch_vmin_value = Float().tag(sync=True)
    stretch_vmin_sync = Dict().tag(sync=True)

    stretch_vmax_value = Float().tag(sync=True)
    stretch_vmax_sync = Dict().tag(sync=True)

    stretch_params_value = Dict().tag(sync=True)
    stretch_params_sync = Dict().tag(sync=True)

    stretch_hist_sync = Dict().tag(sync=True)
    stretch_hist_zoom_limits = Bool().tag(sync=True)
    stretch_hist_nbins = IntHandleEmpty(25).tag(sync=True)
    stretch_histogram_widget = Unicode().tag(sync=True)

    stretch_curve_visible = Bool(True).tag(sync=True)

    subset_visible_value = Bool().tag(sync=True)
    subset_visible_sync = Dict().tag(sync=True)

    subset_color_value = Unicode().tag(sync=True)
    subset_color_sync = Dict().tag(sync=True)

    subset_opacity_value = Float().tag(sync=True)
    subset_opacity_sync = Dict().tag(sync=True)

    image_visible_value = Bool().tag(sync=True)
    image_visible_sync = Dict().tag(sync=True)

    image_color_mode_value = Unicode().tag(sync=True)
    image_color_mode_sync = Dict().tag(sync=True)

    image_color_value = Any().tag(sync=True)
    image_color_sync = Dict().tag(sync=True)

    image_colormap_value = Unicode().tag(sync=True)
    image_colormap_sync = Dict().tag(sync=True)

    image_opacity_value = Float().tag(sync=True)
    image_opacity_sync = Dict().tag(sync=True)

    image_contrast_value = Float().tag(sync=True)
    image_contrast_sync = Dict().tag(sync=True)

    image_bias_value = Float().tag(sync=True)
    image_bias_sync = Dict().tag(sync=True)

    contour_spinner = Bool().tag(sync=True)

    contour_visible_value = Bool().tag(sync=True)
    contour_visible_sync = Dict().tag(sync=True)

    contour_mode_value = Unicode().tag(sync=True)
    contour_mode_sync = Dict().tag(sync=True)

    contour_min_value = Float().tag(sync=True)
    contour_min_sync = Dict().tag(sync=True)

    contour_max_value = Float().tag(sync=True)
    contour_max_sync = Dict().tag(sync=True)

    contour_nlevels_value = Int().tag(sync=True)
    contour_nlevels_sync = Dict().tag(sync=True)

    contour_custom_levels_value = List().tag(sync=True)
    contour_custom_levels_txt = Unicode().tag(sync=True)   # controlled by vue
    contour_custom_levels_sync = Dict().tag(sync=True)

    axes_visible_value = Bool().tag(sync=True)
    axes_visible_sync = Dict().tag(sync=True)

    icon_radialtocheck = Unicode(read_icon(os.path.join(ICON_DIR, 'radialtocheck.svg'), 'svg+xml')).tag(sync=True)  # noqa
    icon_checktoradial = Unicode(read_icon(os.path.join(ICON_DIR, 'checktoradial.svg'), 'svg+xml')).tag(sync=True)  # noqa

    cmap_samples = Dict(cmap_samples).tag(sync=True)
    swatches_palette = List().tag(sync=True)
    apply_RGB_presets_spinner = Bool(False).tag(sync=True)
    stretch_hist_spinner = Bool(False).tag(sync=True)

    volume_value = IntHandleEmpty(50).tag(sync=True)
    volume_sync = Dict().tag(sync=True)

    sonified_audible_value = Bool(False).tag(sync=True)
    sonified_audible_sync = Dict().tag(sync=True)

    hist_visible_value = Bool().tag(sync=True)
    hist_visible_sync = Dict().tag(sync=True)

    hist_opacity_value = Float().tag(sync=True)
    hist_opacity_sync = Dict().tag(sync=True)

    hist_color_value = Any().tag(sync=True)
    hist_color_sync = Dict().tag(sync=True)

    hist_xlog_value = Bool().tag(sync=True)
    hist_xlog_sync = Dict().tag(sync=True)

    hist_ylog_value = Bool().tag(sync=True)
    hist_ylog_sync = Dict().tag(sync=True)

    hist_n_bin_value = Int().tag(sync=True)
    hist_n_bin_sync = Dict().tag(sync=True)

    hist_x_min_value = Float().tag(sync=True)
    hist_x_min_sync = Dict().tag(sync=True)

    hist_x_max_value = Float().tag(sync=True)
    hist_x_max_sync = Dict().tag(sync=True)

    hist_cumulative_value = Bool().tag(sync=True)
    hist_cumulative_sync = Dict().tag(sync=True)

    hist_normalize_value = Bool().tag(sync=True)
    hist_normalize_sync = Dict().tag(sync=True)

    hist_update_bins_on_reset_limits_value = Bool().tag(sync=True)
    hist_update_bins_on_reset_limits_sync = Dict().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Set viewer and layer display options.'

        self.layer = LayerSelect(self, 'layer_items', 'layer_selected',
                                 'viewer_selected', 'layer_multiselect')

        self.layer.filters += [is_not_wcs_only, 'has_wcs_if_image_viewer_pixel_linked']

        self.swatches_palette = [
            ['#FF0000', '#AA0000', '#550000'],
            ['#FFD300', '#AAAA00', '#555500'],
            ['#4CFF00', '#00AA00', '#005500'],
            ['#00FF8E', '#00AAAA', '#005555'],
            ['#0089FF', '#5200FF', '#000055']
        ]

        def is_profile(state):
            return isinstance(state, (ProfileViewerState, ProfileLayerState))

        def not_profile(state):
            return not is_profile(state)

        def is_scatter(state):
            return isinstance(state, (ScatterViewerState, BqplotScatterLayerState))

        def is_histogram(state):
            return isinstance(state, (HistogramViewerState, BqplotHistogramLayerState))

        def is_scatter_or_histogram(state):
            return is_scatter(state) or is_histogram(state)

        def supports_line(state):
            return is_profile(state) or is_scatter(state)

        def is_image(state):
            return isinstance(state, BqplotImageLayerState)

        def not_image(state):
            return not is_image(state)

        def not_image_viewer(state):
            return not isinstance(state, ImageViewerState)

        def not_image_or_spatial_subset(state):
            return not is_image(state) and not is_spatial_subset(state)

        def is_spatial_subset(state):
            return isinstance(state, ImageSubsetLayerState) and is_not_wcs_only(state.layer)

        def is_not_subset(state):
            return not is_spatial_subset(state)

        def is_sonified(state):
            return isinstance(state, SonifiedLayerState)

        def line_visible(state):
            # exclude for scatter layers where the marker is shown instead of the line
            return getattr(state, 'line_visible', True)

        def state_attr_for_line_visible(state):
            if is_scatter(state):
                return 'line_visible'
            return 'visible'

        # Profile/line viewer/layer options:
        self.line_visible = PlotOptionsSyncState(self, self.viewer, self.layer, state_attr_for_line_visible,  # noqa
                                                 'line_visible_value', 'line_visible_sync',
                                                 state_filter=supports_line)
        self.line_color = PlotOptionsSyncState(self, self.viewer, self.layer, 'color',
                                               'line_color_value', 'line_color_sync',
                                               state_filter=not_image_or_spatial_subset)
        self.line_width = PlotOptionsSyncState(self, self.viewer, self.layer, 'linewidth',
                                               'line_width_value', 'line_width_sync',
                                               state_filter=supports_line)
        self.line_opacity = PlotOptionsSyncState(self, self.viewer, self.layer, 'alpha',
                                                 'line_opacity_value', 'line_opacity_sync',
                                                 state_filter=supports_line)
        self.line_as_steps = PlotOptionsSyncState(self, self.viewer, self.layer, 'as_steps',
                                                  'line_as_steps_value', 'line_as_steps_sync')
        self.uncertainty_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'show_uncertainty',  # noqa
                                                        'uncertainty_visible_value', 'uncertainty_visible_sync')  # noqa

        self.xatt = PlotOptionsSyncState(self, self.viewer, self.layer, 'x_att',
                                         'xatt_value', 'xatt_sync',
                                         state_filter=is_scatter_or_histogram)

        self.yatt = PlotOptionsSyncState(self, self.viewer, self.layer, 'y_att',
                                         'yatt_value', 'yatt_sync',
                                         state_filter=is_scatter)

        # Viewer bounds
        self.x_min = PlotOptionsSyncState(self, self.viewer, self.layer, 'x_min',
                                          'x_min_value', 'x_min_sync',
                                          state_filter=not_image_viewer)
        self.x_max = PlotOptionsSyncState(self, self.viewer, self.layer, 'x_max',
                                          'x_max_value', 'x_max_sync',
                                          state_filter=not_image_viewer)
        self.y_min = PlotOptionsSyncState(self, self.viewer, self.layer, 'y_min',
                                          'y_min_value', 'y_min_sync',
                                          state_filter=not_image_viewer)
        self.y_max = PlotOptionsSyncState(self, self.viewer, self.layer, 'y_max',
                                          'y_max_value', 'y_max_sync',
                                          state_filter=not_image_viewer)
        self.zoom_center_x = PlotOptionsSyncState(self, self.viewer, self.layer, 'zoom_center_x',
                                                  'zoom_center_x_value', 'zoom_center_x_sync')
        self.zoom_center_y = PlotOptionsSyncState(self, self.viewer, self.layer, 'zoom_center_y',
                                                  'zoom_center_y_value', 'zoom_center_y_sync')
        self.zoom_radius = PlotOptionsSyncState(self, self.viewer, self.layer, 'zoom_radius',
                                                'zoom_radius_value', 'zoom_radius_sync')

        # Scatter/marker options:
        # NOTE: marker_visible hides the entire layer (including the line)
        self.marker_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'visible',
                                                   'marker_visible_value', 'marker_visible_sync',
                                                   state_filter=is_scatter)
        self.marker_fill = PlotOptionsSyncState(self, self.viewer, self.layer, 'fill',
                                                'marker_fill_value', 'marker_fill_sync',
                                                state_filter=is_scatter)
        self.marker_opacity = PlotOptionsSyncState(self, self.viewer, self.layer, 'alpha',
                                                   'marker_opacity_value', 'marker_opacity_sync',
                                                   state_filter=is_scatter)
        self.marker_size_mode = PlotOptionsSyncState(self, self.viewer, self.layer, 'size_mode',
                                                     'marker_size_mode_value', 'marker_size_mode_sync',  # noqa
                                                     state_filter=is_scatter)
        self.marker_size = PlotOptionsSyncState(self, self.viewer, self.layer, 'size',
                                                'marker_size_value', 'marker_size_sync',
                                                state_filter=is_scatter)
        self.marker_size_scale = PlotOptionsSyncState(self, self.viewer, self.layer, 'size_scaling',
                                                      'marker_size_scale_value', 'marker_size_scale_sync',  # noqa
                                                      state_filter=is_scatter)
        self.marker_size_col = PlotOptionsSyncState(self, self.viewer, self.layer, 'size_att',
                                                    'marker_size_col_value', 'marker_size_col_sync',
                                                    state_filter=is_scatter)
        self.marker_size_vmin = PlotOptionsSyncState(self, self.viewer, self.layer, 'size_vmin',
                                                     'marker_size_vmin_value', 'marker_size_vmin_sync',  # noqa
                                                     state_filter=is_scatter)
        self.marker_size_vmax = PlotOptionsSyncState(self, self.viewer, self.layer, 'size_vmax',
                                                     'marker_size_vmax_value', 'marker_size_vmax_sync',  # noqa
                                                     state_filter=is_scatter)

        # TODO: remove marker_ prefix if these also apply to the lines?
        self.marker_color_mode = PlotOptionsSyncState(self, self.viewer, self.layer, 'cmap_mode',
                                                      'marker_color_mode_value', 'marker_color_mode_sync',  # noqa
                                                      state_filter=is_scatter)
        self.marker_color = PlotOptionsSyncState(self, self.viewer, self.layer, 'color',
                                                 'marker_color_value', 'marker_color_sync',
                                                 state_filter=is_scatter)
        self.marker_color_col = PlotOptionsSyncState(self, self.viewer, self.layer, 'cmap_att',
                                                     'marker_color_col_value', 'marker_color_col_sync',  # noqa
                                                     state_filter=is_scatter)
        self.marker_colormap = PlotOptionsSyncState(self, self.viewer, self.layer, 'cmap',
                                                    'marker_colormap_value', 'marker_colormap_sync',
                                                    state_filter=is_scatter)
        self.marker_colormap_vmin = PlotOptionsSyncState(self, self.viewer, self.layer, 'cmap_vmin',
                                                         'marker_colormap_vmin_value', 'marker_colormap_vmin_sync',  # noqa
                                                         state_filter=is_scatter)
        self.marker_colormap_vmax = PlotOptionsSyncState(self, self.viewer, self.layer, 'cmap_vmax',
                                                         'marker_colormap_vmax_value', 'marker_colormap_vmax_sync',  # noqa
                                                         state_filter=is_scatter)

        # Image viewer/layer options:
        self.stretch_function = PlotOptionsSyncState(self, self.viewer, self.layer, 'stretch',
                                                     'stretch_function_value', 'stretch_function_sync',  # noqa
                                                     state_filter=is_image)
        # use add_observe to ensure that the glue state syncs with the traitlet choice:
        self.stretch_function.add_observe('stretch_function_value', self._update_stretch_curve)

        self.stretch_preset = PlotOptionsSyncState(self, self.viewer, self.layer, 'percentile',
                                                   'stretch_preset_value', 'stretch_preset_sync',
                                                   state_filter=is_image)
        self.stretch_vmin = PlotOptionsSyncState(self, self.viewer, self.layer, 'v_min',
                                                 'stretch_vmin_value', 'stretch_vmin_sync',
                                                 state_filter=is_image)
        self.stretch_vmax = PlotOptionsSyncState(self, self.viewer, self.layer, 'v_max',
                                                 'stretch_vmax_value', 'stretch_vmax_sync',
                                                 state_filter=is_image)
        self.stretch_params = PlotOptionsSyncState(self, self.viewer, self.layer, 'stretch_parameters',  # noqa
                                                   'stretch_params_value', 'stretch_params_sync',
                                                   state_filter=is_image)

        self.subset_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'visible',
                                                   'subset_visible_value', 'subset_visible_sync',
                                                   state_filter=is_spatial_subset)
        self.subset_color = PlotOptionsSyncState(self, self.viewer, self.layer, 'color',
                                                 'subset_color_value', 'subset_color_sync',
                                                 state_filter=is_spatial_subset)
        self.subset_opacity = PlotOptionsSyncState(self, self.viewer, self.layer, 'alpha',
                                                   'subset_opacity_value', 'subset_opacity_sync',
                                                   state_filter=is_spatial_subset)
        self.image_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'bitmap_visible',
                                                  'image_visible_value', 'image_visible_sync',
                                                  state_filter=is_image)
        self.image_color_mode = PlotOptionsSyncState(self, self.viewer, self.layer, 'color_mode',  # noqa
                                                     'image_color_mode_value', 'image_color_mode_sync')  # noqa
        self.image_color = PlotOptionsSyncState(self, self.viewer, self.layer, 'color',
                                                'image_color_value', 'image_color_sync',
                                                state_filter=is_image)
        self.image_colormap = PlotOptionsSyncState(self, self.viewer, self.layer, 'cmap',
                                                   'image_colormap_value', 'image_colormap_sync')
        self.image_opacity = PlotOptionsSyncState(self, self.viewer, self.layer, 'alpha',
                                                  'image_opacity_value', 'image_opacity_sync',
                                                  state_filter=is_image)
        self.image_contrast = PlotOptionsSyncState(self, self.viewer, self.layer, 'contrast',
                                                   'image_contrast_value', 'image_contrast_sync')
        self.image_bias = PlotOptionsSyncState(self, self.viewer, self.layer, 'bias',
                                               'image_bias_value', 'image_bias_sync')

        self.contour_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'contour_visible',  # noqa
                                                    'contour_visible_value', 'contour_visible_sync',
                                                    spinner='contour_spinner')
        self.contour_mode = PlotOptionsSyncState(self, self.viewer, self.layer, 'level_mode',
                                                 'contour_mode_value', 'contour_mode_sync',
                                                 spinner='contour_spinner')
        self.contour_min = PlotOptionsSyncState(self, self.viewer, self.layer, 'c_min',
                                                'contour_min_value', 'contour_min_sync',
                                                spinner='contour_spinner')
        self.contour_max = PlotOptionsSyncState(self, self.viewer, self.layer, 'c_max',
                                                'contour_max_value', 'contour_max_sync',
                                                spinner='contour_spinner')
        self.contour_nlevels = PlotOptionsSyncState(self, self.viewer, self.layer, 'n_levels',
                                                    'contour_nlevels_value', 'contour_nlevels_sync',
                                                    spinner='contour_spinner')
        self.contour_custom_levels = PlotOptionsSyncState(self, self.viewer, self.layer, 'levels',
                                                          'contour_custom_levels_value', 'contour_custom_levels_sync',   # noqa
                                                          spinner='contour_spinner')

        self.volume_level = PlotOptionsSyncState(self, self.viewer, self.layer, 'volume',
                                                 'volume_value', 'volume_sync',
                                                 state_filter=is_sonified)

        self.sonified_audible = PlotOptionsSyncState(self, self.viewer, self.layer, 'audible',
                                                     'sonified_audible_value',
                                                     'sonified_audible_sync',
                                                     state_filter=is_sonified)

        # Histogram layer options:
        self.hist_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'visible',
                                                 'hist_visible_value', 'hist_visible_sync',
                                                 state_filter=is_histogram)

        self.hist_opacity = PlotOptionsSyncState(self, self.viewer, self.layer, 'alpha',
                                                 'hist_opacity_value', 'hist_opacity_sync',
                                                 state_filter=is_histogram)

        self.hist_color = PlotOptionsSyncState(self, self.viewer, self.layer, 'color',
                                               'hist_color_value', 'hist_color_sync',
                                               state_filter=is_histogram)

        # Axes options:
        # axes_visible hidden for imviz in plot_options.vue
        self.axes_visible = PlotOptionsSyncState(self, self.viewer, self.layer, 'show_axes',
                                                 'axes_visible_value', 'axes_visible_sync',
                                                 state_filter=not_profile)
        sv = self.spectrum_viewer
        if sv is not None:
            sv.state.add_callback('x_display_unit',
                                  self._on_global_display_unit_changed)
            sv.state.add_callback('y_display_unit',
                                  self._on_global_display_unit_changed)

        # Histogram axes options:
        self.hist_xlog = PlotOptionsSyncState(self, self.viewer, self.layer, 'x_log',
                                              'hist_xlog_value', 'hist_xlog_sync',
                                              state_filter=is_histogram)
        self.hist_ylog = PlotOptionsSyncState(self, self.viewer, self.layer, 'y_log',
                                              'hist_ylog_value', 'hist_ylog_sync',
                                              state_filter=is_histogram)
        self.hist_n_bin = PlotOptionsSyncState(self, self.viewer, self.layer, 'hist_n_bin',
                                               'hist_n_bin_value', 'hist_n_bin_sync',
                                               state_filter=is_histogram)
        self.hist_x_min = PlotOptionsSyncState(self, self.viewer, self.layer, 'hist_x_min',
                                               'hist_x_min_value', 'hist_x_min_sync',
                                               state_filter=is_histogram)
        self.hist_x_max = PlotOptionsSyncState(self, self.viewer, self.layer, 'hist_x_max',
                                               'hist_x_max_value', 'hist_x_max_sync',
                                               state_filter=is_histogram)
        self.hist_cumulative = PlotOptionsSyncState(self, self.viewer, self.layer, 'cumulative',
                                                    'hist_cumulative_value', 'hist_cumulative_sync',
                                                    state_filter=is_histogram)
        self.hist_normalize = PlotOptionsSyncState(self, self.viewer, self.layer, 'normalize',
                                                   'hist_normalize_value', 'hist_normalize_sync',
                                                   state_filter=is_histogram)
        self.hist_update_bins_on_reset_limits = PlotOptionsSyncState(self, self.viewer, self.layer, 'update_bins_on_reset_limits',  # noqa
                                                                     'hist_update_bins_on_reset_limits_value',  # noqa
                                                                     'hist_update_bins_on_reset_limits_sync',  # noqa
                                                                     state_filter=is_histogram)

        # Add layer callback to image viewers to track active layer
        for viewer in self.app._viewer_store.values():
            viewer.state.add_callback('layers', lambda msg: self._layers_changed(viewer=viewer))

        self.hub.subscribe(self, ViewerAddedMessage, handler=self._on_viewer_added)

        self.hub.subscribe(self, ChangeRefDataMessage,
                           handler=self._on_refdata_change)

        if self.config == 'deconfigged':
            self.observe_traitlets_for_relevancy(traitlets_to_observe=['viewer_items'])

    @property
    def user_api(self):
        expose = ['multiselect', 'viewer', 'viewer_multiselect', 'layer', 'layer_multiselect',
                  'select_all', 'subset_visible', 'reset_viewer_bounds']
        if self.config == "cubeviz":
            expose += ['uncertainty_visible', 'volume_level', 'sonified_audible']
        if self.config != "imviz":
            expose += ['x_min', 'x_max', 'y_min', 'y_max',
                       'axes_visible', 'line_visible', 'line_color', 'line_width', 'line_opacity',
                       'line_as_steps', 'uncertainty_visible']
        if self.config != "specviz":
            expose += ['zoom_center_x', 'zoom_center_y', 'zoom_radius',
                       'subset_color', 'subset_opacity',
                       'stretch_function', 'stretch_preset', 'stretch_vmin', 'stretch_vmax',
                       'stretch_hist_zoom_limits', 'stretch_hist_nbins',
                       'image_visible', 'image_color_mode',
                       'image_color', 'image_colormap', 'image_opacity',
                       'image_contrast', 'image_bias',
                       'contour_visible', 'contour_mode',
                       'contour_min', 'contour_max', 'contour_nlevels', 'contour_custom_levels',
                       'stretch_curve_visible', 'apply_RGB_presets']
        if self.config == 'deconfigged':
            expose += ['xatt', 'yatt', 'hist_visible', 'hist_color', 'hist_opacity',
                       'hist_xlog', 'hist_ylog', 'hist_n_bin', 'hist_x_min', 'hist_x_max',
                       'hist_cumulative', 'hist_normalize', 'hist_update_bins_on_reset_limits']
        if self.config in ('imviz', 'deconfigged'):
            expose += ['marker_visible', 'marker_fill', 'marker_opacity',
                       'marker_size', 'marker_size_scale',
                       'marker_color_mode', 'marker_color',
                       'marker_color_col', 'marker_colormap',
                       'marker_colormap_vmin', 'marker_colormap_vmax']

        return PluginUserApi(self, expose)

    @property
    def multiselect(self):
        warnings.warn(
            "multiselect has been replaced by separate viewer_multiselect and "
            "layer_multiselect and will be removed in the future. "
            "This currently evaluates viewer_multiselect or layer_multiselect", DeprecationWarning)
        return self.viewer_multiselect or self.layer_multiselect

    @multiselect.setter
    def multiselect(self, value):
        warnings.warn(
            "multiselect has been replaced by separate viewer_multiselect and "
            "layer_multiselect and will be removed in the future. "
            "This currently sets viewer_multiselect and layer_multiselect", DeprecationWarning)
        self.viewer_multiselect = value
        self.layer_multiselect = value

    @cached_property
    def stretch_histogram(self):
        stretch_histogram = Plot(self, name='stretch_hist', viewer_type='histogram',
                                 update_callback=self._update_stretch_histogram)
        # Add the stretch bounds tool to the default Plot viewer.
        stretch_histogram.tools_nested.append(["jdaviz:stretch_bounds"])
        stretch_histogram._initialize_toolbar(["jdaviz:stretch_bounds"])

        stretch_histogram._add_data('histogram', x=[0, 1])

        stretch_histogram.add_line('vmin', x=[0, 0], y=[0, 1], ynorm=True, color='#c75d2c')
        stretch_histogram.add_line('vmax', x=[0, 0], y=[0, 1], ynorm='vmin', color='#c75d2c')
        stretch_histogram.add_line(
            label='stretch_curve',
            x=[], y=[],
            ynorm='vmin',
            color="#007BA1",  # "inactive" blue
            opacities=[0.5],
        )
        stretch_histogram.add_scatter(
            label='stretch_knots',
            x=[], y=[],
            ynorm='vmin',
            color="#c75d2c",  # "active" orange (tool enabled by default)
        )
        stretch_histogram.add_scatter('colorbar', x=[], y=[], ynorm='vmin', marker='square', stroke_width=33)  # noqa: E501
        stretch_histogram.viewer.state.update_bins_on_reset_limits = False
        stretch_histogram.viewer.state.x_limits_percentile = 95
        with stretch_histogram.figure.hold_sync():
            stretch_histogram.figure.axes[0].label = 'pixel value'
            stretch_histogram.figure.axes[0].num_ticks = 3
            stretch_histogram.figure.axes[0].tick_format = '0.1e'
            stretch_histogram.figure.axes[1].label = 'density'
            stretch_histogram.figure.axes[1].num_ticks = 2
        self.stretch_histogram_widget = f'IPY_MODEL_{stretch_histogram.model_id}'
        self.send_state('stretch_histogram_widget')
        return stretch_histogram

    def select_all(self, viewers=True, layers=True):
        """
        Enable multiselect mode and select all viewers and/or layers.

        Parameters
        ----------
        viewers : bool
            Whether to set ``viewer_multiselect`` and select all viewers (default: True)

        layers: bool
            Whether to set ``layer_multiselect`` and select all layers (default: True)
        """
        if viewers:
            self.viewer_multiselect = True
            self.viewer.select_all()
        if layers:
            self.layer_multiselect = True
            self.layer.select_all()

    def _on_global_display_unit_changed(self, *args):
        sv = self.spectrum_viewer
        self.display_units['spectral'] = sv.state.x_display_unit
        self.display_units['flux'] = sv.state.y_display_unit
        self.send_state('display_units')

    def _on_refdata_change(self, *args):
        if self.app._align_by.lower() == 'wcs':
            self.display_units['image'] = 'deg'
        else:
            self.display_units['image'] = 'pix'
        self.send_state('display_units')
        self._update_viewer_zoom_steps()

    def _on_viewer_added(self, msg):
        viewer = self.app.get_viewer_by_id(msg.viewer_id)
        viewer.state.add_callback('layers', lambda msg: self._layers_changed(viewer=viewer))

    @observe('viewer_selected')
    def _layers_changed(self, msg=None, viewer=None):
        # We need msg first in the keyword arguments to catch the msg value from the observe,
        # even though we don't end up using it
        if self.viewer_multiselect or not hasattr(self, 'viewer'):
            self.active_layer = ""
            return

        if viewer is None:
            viewer = self.viewer.selected_obj

        if viewer is self.viewer.selected_obj and self._viewer_is_image_viewer():  # noqa
            if viewer.active_image_layer is None:
                self.active_layer = ""
                return
            self.active_layer = viewer.active_image_layer.layer.label

    def vue_unmix_state(self, names):
        if isinstance(names, str):
            names = [names]
        for name in names:
            sync_state = getattr(self, name)
            sync_state.unmix_state()
        if 'stretch_params' in names:
            # there is no way to call send_state to force the update to the layers,
            # so we'll force an update by clearing first
            stretch_params = dict(self.stretch_params_value)
            self.stretch_params_value = {}
            self.stretch_params_value = stretch_params

    def vue_set_value(self, data):
        attr_name = data.get('name')
        value = data.get('value')
        setattr(self, attr_name, value)

    @with_spinner('apply_RGB_presets_spinner')
    def apply_RGB_presets(self):
        """
        Applies preset colors, opacities, and stretch settings to all visible layers
        (in all viewers) when in Color (Monochromatic) mode.
        """

        if (self.image_color_mode_value != "One color per layer" or
                self.image_color_mode_sync['mixed']):
            raise ValueError("RGB presets can only be applied if color mode is Color.")
        # Preselected colors we want to use for 5 or less layers
        preset_colors = [self.swatches_palette[4][1],
                         "#0000FF",
                         "#00FF00",
                         self.swatches_palette[1][0],
                         self.swatches_palette[0][0],
                         ]

        preset_inds = {2: [1, 4], 3: [1, 2, 4], 4: [1, 2, 3, 4]}

        # Switch back to this at the end
        initial_layer = self.layer_selected

        # Determine layers visible in selected viewer(s) - consider mixed to be visible
        visible_layers = [layer['label'] for layer in self.layer.items if not layer['is_subset'] and (layer['visible'] in (True, 'mixed'))]  # noqa

        # Set opacity to something that seems sensible
        n_visible = len(visible_layers)
        default_opacity = 1
        if n_visible > 2:
            default_opacity = 1 / math.log2(n_visible)

        # Sample along a colormap if we have too many layers
        if n_visible > len(preset_colors):
            cmap = matplotlib.colormaps['gist_rainbow'].resampled(n_visible)
            # Have to reverse the order of the cmap to make physical sense with
            # assumed wavelength order of layers.
            preset_colors = [matplotlib.colors.to_hex(cmap(i), keep_alpha=True) for
                             i in range(n_visible - 1, -1, -1)]
        elif n_visible >= 2 and n_visible < len(preset_colors):
            preset_colors = [preset_colors[i] for i in preset_inds[n_visible]]

        for i in range(n_visible):
            self.layer_selected = visible_layers[i]
            self.image_opacity.unmix_state(default_opacity)
            self.image_color.unmix_state(preset_colors[i])
            self.stretch_function.unmix_state("arcsinh")
            self.stretch_preset.unmix_state(99)

        self.layer_selected = initial_layer

    def vue_apply_RGB_presets(self, data):
        self.apply_RGB_presets()

    @observe('viewer_selected',
             'x_min_value', 'x_max_value',
             'y_min_value', 'y_max_value')
    def _update_viewer_bound_steps(self, msg={}):
        if not hasattr(self, 'viewer'):  # pragma: no cover
            # plugin hasn't been fully initialized yet
            return

        if not self.viewer.selected or not self.x_min_sync['in_subscribed_states']:
            # nothing selected yet
            return

        for ax in ('x', 'y'):
            ax_min = getattr(self, f'{ax}_min_value')
            ax_max = getattr(self, f'{ax}_max_value')
            bound_step, decimals = _round_step((ax_max - ax_min) / 100.)
            decimals = -int(np.log10(abs(bound_step))) + 1 if bound_step != 0 else 6
            setattr(self, f'{ax}_bound_step', bound_step)
            setattr(self, f'{ax}_min_value', np.round(ax_min, decimals=decimals))
            setattr(self, f'{ax}_max_value', np.round(ax_max, decimals=decimals))

    @observe('viewer_selected',
             'zoom_center_x_value', 'zoom_center_y_value',
             'zoom_radius_value')
    def _update_viewer_zoom_steps(self, msg={}):
        if not hasattr(self, 'viewer'):  # pragma: no cover
            # plugin hasn't been fully initialized yet
            return

        if not self.viewer.selected or not self.zoom_radius_sync['in_subscribed_states']:
            # nothing selected yet
            return

        # in the case of multiple viewers, calculate based on the first
        # alternatively, we could find the most extreme by looping over all selected viewers
        viewers = self.viewer.selected_obj if self.viewer_multiselect else [self.viewer.selected_obj]  # noqa
        for viewer in viewers:
            if hasattr(viewer.state, '_get_reset_limits'):
                break
        else:
            # no image viewer
            return
        x_min, x_max, y_min, y_max = viewer.state._get_reset_limits(return_as_world=True)
        self.zoom_step, _ = _round_step(max(x_max-x_min, y_max-y_min) / 100.)

    def vue_reset_viewer_bounds(self, _):
        self.reset_viewer_bounds()

    def reset_viewer_bounds(self):
        # This button is currently only exposed if only the spectrum viewer is selected
        viewers = [self.viewer.selected_obj] if not self.viewer_multiselect else self.viewer.selected_obj # noqa
        for viewer in viewers:
            viewer.toolbar.tools['jdaviz:homezoom'].activate()

    @observe('stretch_function_sync', 'stretch_params_sync',
             'stretch_vmin_sync', 'stretch_vmax_sync',
             'image_color_mode_sync', 'image_color_sync', 'image_colormap_sync')
    def _update_stretch_hist_sync(self, msg={}):
        # the histogram should show as mixed if ANY of the input parameters are mixed
        # these should match in the @observe above, all_syncs here, as well as the strings
        # passed to unmix_state in the <glue-state-sync-wrapper> in plot_options.vue
        all_syncs = [self.stretch_function_sync, self.stretch_params_sync,
                     self.stretch_vmin_sync, self.stretch_vmax_sync,
                     self.image_color_mode_sync, self.image_color_sync, self.image_colormap_sync]
        self.stretch_hist_sync = {'in_subscribed_states': bool(np.any([sync.get('in_subscribed_states', False) for sync in all_syncs])),  # noqa
                                  'mixed': bool(np.any([sync.get('mixed', False) for sync in all_syncs]))}  # noqa

    @observe('is_active', 'layer_selected', 'viewer_selected',
             'stretch_hist_zoom_limits')
    @skip_if_no_updates_since_last_active()
    def _request_update_stretch_histogram(self, msg={}):
        if not hasattr(self, 'viewer'):  # pragma: no cover
            # plugin hasn't been fully initialized yet
            return

        # NOTE: this method is separate from _update_stretch_histogram so that
        # _update_stretch_histogram can be called manually (or from the
        # update_callback on the Plot object itself) without going through
        # the skip_if_no_updates_since_last_active check (and can therefore
        # be executed even if the plugin is not active)
        self._update_stretch_histogram(msg)

    def _zoom_limits_update_stretch_histogram(self, msg={}):
        if not self.stretch_hist_zoom_limits:
            # there isn't anything to update, let's not waste resources
            return
        self._update_stretch_histogram()

    @with_spinner('stretch_hist_spinner')
    def _update_stretch_histogram(self, msg={}):
        if not self.stretch_function_sync.get('in_subscribed_states'):  # pragma: no cover
            # no (image) viewer with stretch function options
            return

        if not self.viewer.selected or not self.layer.selected:  # pragma: no cover
            # nothing to plot, will be hidden in UI
            return

        if self.layer_multiselect and len(self.layer.selected) > 1:
            # currently only support single-layer, if multiple layers are selected, the plot
            # will be hidden in the UI
            return

        if not self._viewer_is_image_viewer() or not self._layer_is_image_layer():
            # don't update histogram if selected viewer is not an image viewer:
            return

        viewer = self.viewer.selected_obj[0] if self.viewer_multiselect else self.viewer.selected_obj  # noqa

        # manage viewer zoom limit callbacks
        if ((isinstance(msg, dict) and msg.get('name') == 'viewer_selected')
                or not self.stretch_hist_zoom_limits):
            vs = viewer.state
            for attr in ('x_min', 'x_max', 'y_min', 'y_max'):
                vs.add_callback(attr, self._zoom_limits_update_stretch_histogram)
        if isinstance(msg, dict) and msg.get('name') == 'viewer_selected':
            viewer_label_old = msg.get('old')
            if isinstance(viewer_label_old, list):
                viewer_label_old = viewer_label_old[0]
            # If the previously selected viewer was deleted, we don't need to do this.
            if viewer_label_old in self.app._viewer_store:
                vs_old = self.app.get_viewer(viewer_label_old).state
                for attr in ('x_min', 'x_max', 'y_min', 'y_max'):
                    vs_old.remove_callback(attr, self._zoom_limits_update_stretch_histogram)

        if not len(self.layer.selected_obj):
            # skip further updates if no data are available:
            return
        if isinstance(self.layer.selected_obj[0], list):
            if not len(self.layer.selected_obj[0]):
                return
            # multiselect case (but we won't check multiselect since the selection can lag behind)
            layer = self.layer.selected_obj[0][0]
        else:
            layer = self.layer.selected_obj[0]
        data = layer.layer

        if isinstance(data, GroupedSubset):
            # don't update histogram for subsets:
            return

        comp = data.get_component(layer.state.attribute)

        # TODO: further optimization could be done by caching sub_data
        if self.stretch_hist_zoom_limits and (not self.layer_multiselect or len(self.layer_selected) == 1):  # noqa
            if hasattr(viewer, '_get_zoom_limits'):
                # Viewer limits. This takes account of Imviz linking.
                xy_limits = viewer._get_zoom_limits(data).astype(int)
                x_limits = xy_limits[:, 0]
                y_limits = xy_limits[:, 1]
                x_min = max(x_limits.min(), 0)
                x_max = x_limits.max()
                y_min = max(y_limits.min(), 0)
                y_max = y_limits.max()

                sub_data = comp.data[y_min:y_max, x_min:x_max]

            else:
                # spectrum-2d-viewer, for example.  We'll assume the viewer
                # limits correspond to the fixed data components from glue
                # and filter directly.
                x_data = data.get_component(data.components[1]).data
                y_data = data.get_component(data.components[0]).data

                inverted_x = getattr(viewer, 'inverted_x_axis', False)
                x_min = viewer.state.x_min if not inverted_x else viewer.state.x_max
                x_max = viewer.state.x_max if not inverted_x else viewer.state.x_min
                inds = np.where((x_data >= x_min) &
                                (x_data <= x_max) &
                                (y_data >= viewer.state.y_min) &
                                (y_data <= viewer.state.y_max))

                sub_data = comp.data[inds]

        else:
            # include all data, regardless of zoom limits
            sub_data = comp.data

        self.stretch_histogram.viewer.state.random_subset = RANDOM_SUBSET_SIZE
        self.stretch_histogram._update_data('histogram', x=sub_data)

        if len(sub_data) > 0:

            # Use glue to compute the statistics since this allows us to use
            # a random subset of the data to compute the histogram.
            # The 2.5 and 97.5 hardcoded here is equivalent to
            # PercentileInterval(95).get_limits(sub_data)
            glue_data = self.stretch_histogram.app.data_collection['histogram']
            hist_lims = (
                glue_data.compute_statistic('percentile', glue_data.id['x'],
                                            percentile=2.5, random_subset=RANDOM_SUBSET_SIZE),
                glue_data.compute_statistic('percentile', glue_data.id['x'],
                                            percentile=97.5, random_subset=RANDOM_SUBSET_SIZE)
            )

            # set the stepsize for vmin/vmax to be approximately 1% of the range of the
            # histogram (within the percentile interval), rounded to 1-2 significant digits
            # to avoid random step sizes.  This logic is somewhat arbitrary and can be safely
            # modified or eventually exposed to the user if that would be useful.
            stretch_vstep = (hist_lims[1] - hist_lims[0]) / 100.
            self.stretch_vstep = _round_step(stretch_vstep)[0]

            with delay_callback(self.stretch_histogram.viewer.state, 'hist_x_min', 'hist_x_max'):
                self.stretch_histogram.viewer.state.hist_x_min = hist_lims[0]
                self.stretch_histogram.viewer.state.hist_x_max = hist_lims[1]

        self.stretch_histogram.figure.title = f"{len(sub_data)} pixels"

        # update the n_bins since this may be a new layer
        self._histogram_nbins_changed()
        # update the curve/colorbar
        self._update_stretch_curve(msg)

    @observe('image_color_mode_value', 'image_color_value', 'image_colormap_value',
             'image_contrast_value', 'image_bias_value',
             'stretch_hist_nbins',
             'stretch_curve_visible',
             'stretch_function_value', 'stretch_vmin_value', 'stretch_vmax_value',
             'stretch_params_value', 'stretch_preset_value',
             'layer_multiselect'
             )
    @skip_if_no_updates_since_last_active()
    def _update_stretch_curve(self, msg=None):
        if not self._viewer_is_image_viewer() or not hasattr(self, 'stretch_histogram'):
            # don't update histogram if selected viewer is not an image viewer,
            # or the stretch histogram hasn't been initialized:
            return

        if self.layer_multiselect and len(self.layer.selected) > 1:
            # currently only support single-layer, if multiple layers are selected, the plot
            # will be hidden in the UI
            return

        # could be multi or single-viewer and/or multi-layer with a single entry,
        # either way, we act on the first entry
        layer = self.layer.selected_obj[0]
        while isinstance(layer, list):
            if not len(layer):
                return
            layer = layer[0]

        if isinstance(layer.layer, GroupedSubset):
            # don't update histogram for subsets, will be hidden in UI
            return

        # create the new/updated stretch curve following the colormapping
        # procedure in glue's CompositeArray:
        interval = ManualInterval(self.stretch_vmin_value, self.stretch_vmax_value)
        contrast_bias = ContrastBiasStretch(self.image_contrast_value, self.image_bias_value)
        stretch = layer.state.stretch_object
        layer_cmap = layer.state.cmap

        # show the colorbar
        color_mode = self.image_color_mode_value

        # NOTE: Index 0 in marks is assumed to be the bin centers.
        x = self.stretch_histogram.figure.marks[0].x
        y = np.ones_like(x)

        # Copied from the __call__ internals of glue/viewers/image/composite_array.py
        data = interval(x)
        data = contrast_bias(data, out=data)
        data = stretch(data, out=data)

        if color_mode == 'Colormaps':
            cmap = glue_colormaps[self.image_colormap.text]
            if hasattr(cmap, "get_bad"):
                bad_color = cmap.get_bad().tolist()[:3]
                layer_cmap = cmap.with_extremes(bad=bad_color + [self.image_opacity_value])
            else:
                layer_cmap = cmap

            # Compute colormapped image
            plane = layer_cmap(data)

        else:  # Color (Monochromatic)
            if self.image_color_value is None:
                # do not crash if image_color_value is not yet assigned,
                # _update_stretch_curve observes image_color_value so will get called again
                return False
            # Get color
            color = COLOR_CONVERTER.to_rgba_array(self.image_color_value)[0]
            plane = data[:, np.newaxis] * color
            plane[:, 3] = 1

        plane = np.clip(plane, 0, 1, out=plane)
        ipycolors = [matplotlib.colors.rgb2hex(p, keep_alpha=False) for p in plane]

        colorbar_mark = self.stretch_histogram.marks['colorbar']
        colorbar_mark.x = x
        colorbar_mark.y = y
        colorbar_mark.colors = ipycolors

        # show "knot" locations if the stretch_function is a spline
        if isinstance(stretch, SplineStretch) and self.stretch_curve_visible:
            knot_mark = self.stretch_histogram.marks['stretch_knots']
            knot_mark.x = (self.stretch_vmin_value +
                           np.asarray(stretch._x) * (self.stretch_vmax_value - self.stretch_vmin_value))  # noqa
            # scale to 0.9 so always falls below colorbar (same as for stretch_curve)
            knot_mark.y = 0.9 * np.asarray(stretch._y)
        else:
            self.stretch_histogram.clear_marks('stretch_knots')

        if self.stretch_curve_visible:
            # create a photoshop style "curve" for the stretch function
            curve_x = np.linspace(self.stretch_vmin_value, self.stretch_vmax_value, 50)
            curve_y = interval(curve_x)
            curve_y = contrast_bias(curve_y)
            curve_y = stretch(curve_y)

            curve_mark = self.stretch_histogram.marks['stretch_curve']
            curve_mark.x = curve_x
            curve_mark.y = 0.9 * curve_y
        else:
            self.stretch_histogram.clear_marks('stretch_curve')

        self.stretch_histogram._refresh_marks()

    @observe('stretch_vmin_value')
    def _stretch_vmin_changed(self, msg=None):
        self.stretch_histogram.marks['vmin'].x = [self.stretch_vmin_value, self.stretch_vmin_value]

    @observe('stretch_vmax_value')
    def _stretch_vmax_changed(self, msg=None):
        self.stretch_histogram.marks['vmax'].x = [self.stretch_vmax_value, self.stretch_vmax_value]

    @observe("stretch_hist_nbins")
    def _histogram_nbins_changed(self, msg={}):
        if self.stretch_histogram is None:
            return
        if self.stretch_hist_nbins == '' or self.stretch_hist_nbins < 1:
            return
        self.stretch_histogram.viewer.state.hist_n_bin = self.stretch_hist_nbins
        # for some reason, this resets the internal marks, so we need to ensure the manual
        # marks are still plotted
        self.stretch_histogram._refresh_marks()

    def set_histogram_limits(self, x_min=None, x_max=None, y_min=None, y_max=None):
        # NOTE: leaving this out of user API until API is finalized with interactive setting
        self.stretch_histogram.set_limits(x_min=x_min, x_max=x_max,
                                          y_min=y_min, y_max=y_max)

    def _viewer_is_image_viewer(self):
        # Import here to prevent circular import (and not at the top of the method so the import
        # check is avoided, whenever possible).
        from jdaviz.configs.imviz.plugins.viewers import ImvizImageView
        from jdaviz.configs.cubeviz.plugins.viewers import CubevizImageView
        from jdaviz.configs.mosviz.plugins.viewers import MosvizImageView, MosvizProfile2DView
        from jdaviz.configs.rampviz.plugins.viewers import RampvizImageView

        def _is_image_viewer(viewer):
            return isinstance(viewer, (ImvizImageView, CubevizImageView,
                                       MosvizImageView, MosvizProfile2DView,
                                       RampvizImageView))

        viewers = self.viewer.selected_obj
        if not isinstance(viewers, list):
            viewers = [viewers]

        return np.all([_is_image_viewer(viewer) for viewer in viewers])

    def _layer_is_image_layer(self):
        from glue_jupyter.bqplot.image.state import BqplotImageLayerState
        return np.all([isinstance(lyr.state, BqplotImageLayerState)
                       for lyr in self.layer.selected_obj_flattened])

    def image_segmentation_map_presets(self, *args, **kwargs):
        # if 'Random' colormap is used for visualizing image segmentation,
        # ensure the stretch limits are the min and max, the stretch function
        # is linear, the contrast is 1.0, and the bias is 0.5. This ensures
        # that all label colors are unique:
        if self.image_colormap_value != 'Random':
            return

        self.stretch_preset.value = 100
        self.stretch_function.value = 'linear'
        self.image_contrast_value = 1
        self.image_bias_value = 0.5

    def vue_image_segmentation_map_presets(self, *args, **kwargs):
        self.image_segmentation_map_presets(*args, **kwargs)



from jdaviz.app import Application
from jdaviz.core.loaders.resolvers.resolver import BaseResolver, find_closest_polygon_mark
from jdaviz.core.marks import RegionOverlay

import numpy as np
from astropy.table import Table


# Create a minimal test class that mimics the resolver behavior
# without this we get an error when attempting to use BaseResolver directly
class TestBaseResolver(BaseResolver):
    template = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def test_server_is_remote_callback():
    # Create app instance
    app = Application()

    # Test the sync
    test_obj = TestBaseResolver(app=app)
    settings = app.state.settings

    # Check default
    assert settings.get('server_is_remote') is False
    assert settings.get('server_is_remote') == test_obj.server_is_remote

    settings['server_is_remote'] = True
    assert settings.get('server_is_remote') == test_obj.server_is_remote

    # Ensure setting test_obj.server_is_remote does not backpropagate
    # (this behavior could change)
    test_obj.server_is_remote = False
    assert settings.get('server_is_remote') != test_obj.server_is_remote


def test_footprint_workflow(deconfigged_helper, image_nddata_wcs):
    deconfigged_helper.load(image_nddata_wcs, format='Image', data_label='test_image')

    table = Table()
    table['Dataset'] = ['obs1', 'obs2', 'obs3']
    table['s_region'] = [
        'POLYGON 337.51 -20.84 337.52 -20.84 337.52 -20.83 337.51 -20.83',
        'POLYGON 337.52 -20.84 337.53 -20.84 337.53 -20.83 337.52 -20.83',
        'POLYGON 337.53 -20.84 337.54 -20.84 337.54 -20.83 337.53 -20.83',
    ]

    ldr = deconfigged_helper.loaders['object']
    ldr.object = table
    ldr.treat_table_as_query = True

    assert ldr._obj.parsed_input_is_query is True
    assert ldr._obj.observation_table_populated is True
    assert 's_region' in ldr._obj.observation_table.headers_avail

    ldr._obj.vue_link_by_wcs()
    assert ldr._obj.is_wcs_linked is True

    ldr._obj.toggle_custom_toolbar()
    assert ldr._obj.custom_toolbar_enabled is True

    # Check footprints in viewer
    viewer = list(deconfigged_helper.app._viewer_store.values())[0]
    footprints = [m for m in viewer.figure.marks if isinstance(m, RegionOverlay)]
    assert len(footprints) == 3
    assert sorted([m.label for m in footprints]) == [0, 1, 2]

    mark = footprints[1]
    px, py = np.mean(mark.x), np.mean(mark.y)
    idx = find_closest_polygon_mark(px, py, footprints)
    assert idx is not None
    assert 0 <= idx < 3

    ldr._obj.toggle_custom_toolbar()
    assert ldr._obj.custom_toolbar_enabled is False
    footprints = [m for m in viewer.figure.marks if isinstance(m, RegionOverlay)]
    assert len(footprints) == 0


def test_remove_footprints(deconfigged_helper, image_nddata_wcs):
    deconfigged_helper.load(image_nddata_wcs, format='Image', data_label='test_image')

    table = Table()
    table['Dataset'] = ['obs1']
    table['s_region'] = [
        'POLYGON 337.499 -20.831 337.501 -20.831 337.501 -20.829 337.499 -20.829'
    ]

    ldr = deconfigged_helper.loaders['object']
    ldr.object = table
    ldr.treat_table_as_query = True
    ldr._obj.vue_link_by_wcs()

    # Display footprints
    ldr._obj.toggle_custom_toolbar()
    viewer = list(deconfigged_helper.app._viewer_store.values())[0]
    footprints = [m for m in viewer.figure.marks if isinstance(m, RegionOverlay)]
    assert len(footprints) == 1

    # Remove footprints
    ldr._obj.toggle_custom_toolbar()

    # Assert marks are removed
    footprints = [m for m in viewer.figure.marks if isinstance(m, RegionOverlay)]
    assert len(footprints) == 0


def test_multiselect(deconfigged_helper, image_nddata_wcs):
    deconfigged_helper.load(image_nddata_wcs, format='Image', data_label='test_image')

    table = Table()
    table['Dataset'] = ['obs1', 'obs2', 'obs3']
    table['s_region'] = [
        'POLYGON 337.499 -20.831 337.501 -20.831 337.501 -20.829 337.499 -20.829',
        'POLYGON 337.502 -20.831 337.504 -20.831 337.504 -20.829 337.502 -20.829',
        'POLYGON 337.505 -20.831 337.507 -20.831 337.507 -20.829 337.505 -20.829'
    ]

    ldr = deconfigged_helper.loaders['object']
    ldr.object = table
    ldr.treat_table_as_query = True
    ldr._obj.vue_link_by_wcs()
    ldr._obj.toggle_custom_toolbar()

    viewer = list(deconfigged_helper.app._viewer_store.values())[0]
    footprints = [m for m in viewer.figure.marks if isinstance(m, RegionOverlay)]
    assert len(footprints) == 3

    # Select observations 0 and 2 via the table
    ldr._obj.observation_table.select_rows([0, 2])

    # Check that selected marks match table selection
    selected_labels = sorted([m.label for m in footprints if m.selected])
    assert selected_labels == [0, 2]


def test_display_valid_footprints(deconfigged_helper, image_nddata_wcs):
    deconfigged_helper.load(image_nddata_wcs, format='Image', data_label='test_image')

    table = Table()
    table['Dataset'] = ['obs1', 'obs2']
    table['s_region'] = [
        'POLYGON 337.499 -20.831 337.501 -20.831 337.501 -20.829 337.499 -20.829',
        'POLYGON 337.502 -20.831 337.504 -20.831 337.504 -20.829 337.502 -20.829'
    ]
    ldr = deconfigged_helper.loaders['object']
    ldr.object = table
    ldr.treat_table_as_query = True
    ldr._obj.vue_link_by_wcs()
    ldr._obj.toggle_custom_toolbar()

    # Assert marks were added
    viewer = list(deconfigged_helper.app._viewer_store.values())[0]
    footprints = [m for m in viewer.figure.marks if isinstance(m, RegionOverlay)]
    assert len(footprints) == 2

    # Verify marks have labels/indices
    for mark in footprints:
        assert isinstance(mark, RegionOverlay)
        assert mark.label in [0, 1]


def test_no_image_data_disables_toolbar(deconfigged_helper):
    table = Table()
    table['Dataset'] = ['obs1']
    table['s_region'] = [
        'POLYGON 337.499 -20.831 337.501 -20.831 337.501 -20.829 337.499 -20.829'
    ]

    ldr = deconfigged_helper.loaders['object']
    ldr.object = table
    ldr.treat_table_as_query = True

    # Although we have footprint data, toolbar should not be enabled
    # because there's no image data to link
    assert len(deconfigged_helper.app.data_collection) == 0
    assert ldr._obj.observation_table_populated is True
    assert ldr._obj.image_data_loaded is False


def test_footprint_with_image_deconfigged(deconfigged_helper, image_nddata_wcs):
    deconfigged_helper.load(image_nddata_wcs, format='Image', data_label='Test Image')
    assert len(deconfigged_helper.app.data_collection) == 1

    table = Table()
    table['Dataset'] = ['obs1']
    table['s_region'] = [
        'POLYGON 337.499 -20.831 337.501 -20.831 337.501 -20.829 337.499 -20.829'
    ]

    ldr = deconfigged_helper.loaders['object']
    ldr.object = table
    ldr.treat_table_as_query = True

    assert ldr._obj.parsed_input_is_query is True
    assert ldr._obj.observation_table_populated is True
    assert ldr._obj.image_data_loaded is True

    ldr._obj.vue_link_by_wcs()
    assert ldr._obj.is_wcs_linked is True

    ldr._obj.toggle_custom_toolbar()
    assert ldr._obj.custom_toolbar_enabled is True
    viewer = list(deconfigged_helper.app._viewer_store.values())[0]
    footprints = [m for m in viewer.figure.marks if isinstance(m, RegionOverlay)]
    assert len(footprints) == 1
    assert footprints[0].label == 0


def test_multiviewer_footprints(deconfigged_helper, image_nddata_wcs):
    deconfigged_helper.load(image_nddata_wcs, format='Image', data_label='Image 1')
    viewer = list(deconfigged_helper.app._viewer_store.values())[0]

    table = Table()
    table['Dataset'] = ['obs1', 'obs2']
    table['s_region'] = [
        'POLYGON 337.499 -20.831 337.501 -20.831 337.501 -20.829 337.499 -20.829',
        'POLYGON 337.502 -20.831 337.504 -20.831 337.504 -20.829 337.502 -20.829'
    ]

    ldr = deconfigged_helper.loaders['object']
    ldr.object = table
    ldr.treat_table_as_query = True
    ldr._obj.vue_link_by_wcs()
    ldr._obj.toggle_custom_toolbar()

    viewer_footprints = [m for m in viewer.figure.marks if isinstance(m, RegionOverlay)]
    assert len(viewer_footprints) == 2
    # Assert labels match
    viewer_labels = sorted([m.label for m in viewer_footprints])
    assert viewer_labels == [0, 1]


def test_multiviewer_selection_sync(deconfigged_helper, image_nddata_wcs):
    deconfigged_helper.load(image_nddata_wcs, format='Image', data_label='Image 1')
    viewer = list(deconfigged_helper.app._viewer_store.values())[0]

    table = Table()
    table['Dataset'] = ['obs1', 'obs2', 'obs3']
    table['s_region'] = [
        'POLYGON 337.499 -20.831 337.501 -20.831 337.501 -20.829 337.499 -20.829',
        'POLYGON 337.502 -20.831 337.504 -20.831 337.504 -20.829 337.502 -20.829',
        'POLYGON 337.505 -20.831 337.507 -20.831 337.507 -20.829 337.505 -20.829'
    ]

    ldr = deconfigged_helper.loaders['object']
    ldr.object = table
    ldr.treat_table_as_query = True
    ldr._obj.vue_link_by_wcs()
    ldr._obj.toggle_custom_toolbar()

    # Select observation in table
    ldr._obj.observation_table.select_rows([0, 2])

    # Check selection is synced in viewer
    footprints = [m for m in viewer.figure.marks if isinstance(m, RegionOverlay)]
    selected_labels = [m.label for m in footprints if m.selected]
    assert sorted(selected_labels) == [0, 2]


def test_add_footprints_to_viewer_method(deconfigged_helper, image_nddata_wcs):
    deconfigged_helper.load(image_nddata_wcs, format='Image', data_label='Image 1')

    table = Table()
    table['Dataset'] = ['obs1']
    table['s_region'] = [
        'POLYGON 337.499 -20.831 337.501 -20.831 337.501 -20.829 337.499 -20.829'
    ]

    ldr = deconfigged_helper.loaders['object']
    ldr.object = table
    ldr.treat_table_as_query = True
    ldr._obj.vue_link_by_wcs()

    # Manually call _add_footprints_to_viewer
    viewer = list(deconfigged_helper.app._viewer_store.values())[0]
    ldr._obj.custom_toolbar_enabled = True
    ldr._obj._add_footprints_to_viewer(viewer)

    # Assert footprints were added
    viewer_footprints = [m for m in viewer.figure.marks if isinstance(m, RegionOverlay)]
    assert len(viewer_footprints) == 1
    assert viewer_footprints[0].label == 0


def test_remove_footprints_from_viewer(deconfigged_helper, image_nddata_wcs):
    deconfigged_helper.load(image_nddata_wcs, format='Image', data_label='Image 1')
    viewer = list(deconfigged_helper.app._viewer_store.values())[0]

    table = Table()
    table['Dataset'] = ['obs1']
    table['s_region'] = [
        'POLYGON 337.499 -20.831 337.501 -20.831 337.501 -20.829 337.499 -20.829'
    ]

    ldr = deconfigged_helper.loaders['object']
    ldr.object = table
    ldr.treat_table_as_query = True
    ldr._obj.vue_link_by_wcs()
    ldr._obj.toggle_custom_toolbar()

    # Assert viewer has footprints
    assert any(isinstance(m, RegionOverlay) for m in viewer.figure.marks)
    # Disable toolbar
    ldr._obj.toggle_custom_toolbar()
    # Assert footprints removed
    assert not any(isinstance(m, RegionOverlay) for m in viewer.figure.marks)



import numpy as np
from contextlib import contextmanager
import multiprocessing as mp
import sys
import os

try:
    from strauss.sonification import Sonification
    from strauss.sources import Events
    from strauss.score import Score
    from strauss.generator import Spectralizer
except ImportError:
    pass

from jdaviz.configs.default.plugins.model_fitting.fitting_backend import generate_spaxel_list
from jdaviz.utils import parallelize_calculation

#  smallest fraction of the max audio amplitude that can be represented by a 16-bit signed integer
INT_MAX = 2**15 - 1
MINVOL = 1/INT_MAX


@contextmanager
def suppress_stderr():
    with open(os.devnull, "w") as devnull:
        old_stderr = sys.stderr
        sys.stderr = devnull
        try:
            yield
        finally:
            sys.stderr = old_stderr


def sonify_spectrum(spec, duration, overlap=0.05, system='mono', srate=44100, fmin=40, fmax=1300,
                    eln=False):
    notes = [["A2"]]
    score = Score(notes, duration)
    # set up spectralizer generator
    generator = Spectralizer(samprate=srate)

    # Lets pick the mapping frequency range for the spectrum...
    generator.modify_preset({'min_freq': fmin, 'max_freq': fmax,
                             'fit_spec_multiples': False,
                             'interpolation_type': 'preserve_power',
                             'equal_loudness_normalisation': eln})

    data = {'spectrum': [spec], 'pitch': [1]}

    # set range in spectral flux representing the maximum and minimum sound frequency power:
    # 0 (numeric): absolute 0 in flux units, such that any flux above 0 will sound.
    # '100' (string): 100th percentile (i.e. maximum value) in spectral flux.
    lims = {'spectrum': (0, '%100')}

    # set up source
    sources = Events(data.keys())
    sources.fromdict(data)
    sources.apply_mapping_functions(map_lims=lims)

    # render and play sonification!
    soni = Sonification(score, sources, generator, system, samprate=srate)
    soni.render()
    soni._make_seamless(overlap)

    return soni.loop_channels['0'].values


class CubeListenerData:
    def __init__(self, cube, wlens, samplerate=44100, duration=1, overlap=0.05, buffsize=1024,
                 bdepth=16, wl_unit=None, audfrqmin=50, audfrqmax=1500, eln=False, vol=None,
                 spectral_axis_index=2, n_cpu=None):
        self.siglen = int(samplerate*(duration-overlap))
        self.cube = cube
        self.dur = duration
        self.bdepth = bdepth
        self.srate = samplerate
        self.maxval = pow(2, bdepth-1) - 1
        self.fadedx = 0
        if n_cpu is None:
            self.n_cpu = mp.cpu_count() - 1
        else:
            self.n_cpu = n_cpu
        # Set spectral axis and spatial axes indices for later use
        self.spectral_axis_index = spectral_axis_index
        spatial_inds = [0, 1, 2]
        spatial_inds.remove(self.spectral_axis_index)
        self.spatial_inds = spatial_inds

        if vol is None:
            self.atten_level = 1
        else:
            self.atten_level = int(np.clip((vol/100)**2, MINVOL, 1))

        self.wl_unit = wl_unit
        self.wlens = wlens

        # control fades
        fade = np.linspace(0, 1, buffsize+1)
        self.ifade = fade[:-1]
        self.ofade = fade[::-1][:-1]

        # mapping frequency limits in Hz
        self.audfrqmin = audfrqmin
        self.audfrqmax = audfrqmax

        # do we normalise for equal loudness?
        self.eln = eln

        self.cbuff = False
        self.cursig = np.zeros(self.siglen, dtype='int16')
        self.newsig = np.zeros(self.siglen, dtype='int16')

        # ensure sigcube isn't too big before we initialise it
        slices = [slice(None),]*3
        slices[spectral_axis_index] = 0
        if self.cube[*slices].size * self.siglen * 2 * pow(1024, -3) > 2:
            raise Exception("Cube projected to be > 2Gb!")

        sigcube_shape = list(self.cube.shape)
        sigcube_shape[spectral_axis_index] = self.siglen
        self.sigcube = np.zeros(sigcube_shape, dtype='int16')

    def set_wl_bounds(self, w1, w2):
        """
        set the wavelength bounds for indexing spectra
        """
        wsrt = np.sort([w1, w2])
        self.wl_bounds = tuple(wsrt)

    def sonify_cube(self):
        """
        Iterate through the cube, convert each spectrum to a signal, and store
        in class attributes
        """
        lo2hi = self.wlens.argsort()[::-1]

        spaxels = generate_spaxel_list(self.cube, self.spectral_axis_index)

        # Callback to collect results from workers into the cubes
        def collect_result(results):
            for i in range(len(results['x'])):
                x = results['x'][i]
                y = results['y'][i]
                sig = results['sig'][i]

                # Store fitted values
                if self.spectral_axis_index in [2, -1]:
                    self.sigcube[x, y, :] = sig
                elif self.spectral_axis_index == 0:
                    self.sigcube[:, y, x] = sig

        # Workers for the parallelization pool.
        workers = (SonifySpaxelWorker(self.cube, spx, lo2hi, self.dur, self.srate,
                                      self.audfrqmin, self.audfrqmax, self.eln, self.maxval,
                                      spectral_axis_index=self.spectral_axis_index)
                   for spx in np.array_split(spaxels, self.n_cpu))

        parallelize_calculation(workers, collect_result, n_cpu=self.n_cpu)

        if self.spectral_axis_index == 2:
            self.cursig[:] = self.sigcube[0, 0, :]
        elif self.spectral_axis_index == 0:
            self.cursig[:] = self.sigcube[:, 0, 0]
        self.newsig[:] = self.cursig[:]

    def player_callback(self, outdata, frames, time, status):
        cur = self.cursig
        new = self.newsig
        sdx = int(time.outputBufferDacTime*self.srate)
        dxs = np.arange(sdx, sdx+frames).astype(int) % self.sigcube.shape[self.spectral_axis_index]
        if self.cbuff:
            outdata[:, 0] = (cur[dxs] * self.ofade).astype('int16')
            outdata[:, 0] += (new[dxs] * self.ifade).astype('int16')
            self.cursig[:] = self.newsig[:]
            self.cbuff = False
        else:
            outdata[:, 0] = self.cursig[dxs]
        outdata[:, 0] //= self.atten_level


class SonifySpaxelWorker:
    """
    A class with callable instances that perform fitting over a
    spaxel. It provides the callable for the `Pool.apply_async`
    function, and also holds everything necessary to perform the
    fit over one spaxel.

    Additionally, the callable computes the realization of the
    model just fitted, over that same spaxel. We cannot do these
    two steps (fit and compute) separately, since we cannot
    modify parameter values in an already built CompoundModel
    instance. We need to use the current model instance while
    it still exists.
    """
    def __init__(self, flux_cube, spaxel_set, lo2hi, dur, srate, audfrqmin, audfrqmax,
                 eln, maxval, spectral_axis_index=2):
        self.cube = flux_cube
        self.spaxel_set = spaxel_set
        self.lo2hi = lo2hi
        self.dur = dur
        self.srate = srate
        self.audfrqmin = audfrqmin
        self.audfrqmax = audfrqmax
        self.eln = eln
        self.maxval = maxval
        self.spectral_axis_index = spectral_axis_index

    def __call__(self):
        results = {'x': [], 'y': [], 'sig': []}

        for spaxel in self.spaxel_set:
            x = spaxel[0]
            y = spaxel[1]
            if self.spectral_axis_index in [2, -1]:
                flux = self.cube[x, y, self.lo2hi]
            elif self.spectral_axis_index == 0:
                flux = self.cube[self.lo2hi, y, x]

            if flux.any():
                sig = sonify_spectrum(flux, self.dur,
                                      srate=self.srate,
                                      fmin=self.audfrqmin,
                                      fmax=self.audfrqmax,
                                      eln=self.eln)
                sig = (sig*self.maxval).astype('int16')
            else:
                continue

            results['x'].append(x)
            results['y'].append(y)
            results['sig'].append(sig)

        return results



import numpy as np
from astropy.nddata import NDDataArray

from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.loaders.importers import BaseImporterToDataCollection


__all__ = ['RampIntegrationImporter']


@loader_importer_registry('Ramp Integration')
class RampIntegrationImporter(BaseImporterToDataCollection):
    template_file = __file__, "../to_dc_with_label.vue"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.config == 'rampviz':
            self.viewer.selected = ['integration-viewer']
            self.viewer.create_new_items = []
        else:
            self.viewer.select_default()

    @staticmethod
    def _get_supported_viewers():
        return [{'label': 'Ramp Integration', 'reference': 'rampviz-profile-viewer'}]

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'rampviz'):
            # NOTE: temporary during deconfig process
            return False

        return isinstance(self.input, (np.ndarray, NDDataArray))

    @property
    def output(self):
        """
        Returns the ramp integration data in a format that can be directly consumed by glue.

        The input is expected to be an NDDataArray with shape (1, 1, n_groups) where:
        - The first two dimensions are spatial (collapsed to 1x1)
        - The third dimension is the group/resultant axis

        This matches the format produced by ramp_extraction.py's extract() method when
        passed to add_results_from_plugin().
        """
        data = self.input

        # If input is a plain numpy array, wrap it in an NDDataArray
        if isinstance(data, np.ndarray):
            data = NDDataArray(data)

        # Ensure the data has metadata (required for glue integration)
        if not hasattr(data, 'meta') or data.meta is None:
            data.meta = {}

        return data

    def __call__(self):
        viewers = self.viewer.selected_obj
        super().__call__()
        if viewers is not None:
            for viewer in viewers:
                viewer._initialize_x_axis()

    def assign_component_type(self, comp_id, comp, units, physical_type):
        if comp_id == 'data':
            physical_type = 'ramp_integration'
        elif comp_id == 'Pixel Axis 2 [x]':
            physical_type = 'ramp_integration_group'
        return super().assign_component_type(comp_id, comp, units, physical_type)



import numpy as np
import pytest
import re
from unittest.mock import patch

from astropy import units as u
from astropy.nddata import StdDevUncertainty

from specutils import Spectrum, SpectrumList

from jdaviz.core.loaders.importers.spectrum_list.spectrum_list import (
    SpectrumListImporter,
    combine_lists_to_1d_spectrum
)

from jdaviz.conftest import FakeSpectrumListImporter, FakeSpectrumListConcatenatedImporter
from jdaviz.utils import create_data_hash


def extract_wfss_info(spec):
    """
    Copied and pasted from spectrum_list.py for testing purposes. If the parsing changes
    then this will need to be updated.
    """
    header = spec.meta.get('header', {})
    exp_num = header.get('EXPGRPID', '0_0_0').split('_')[-2]
    source_id = str(spec.meta.get('source_id', ''))
    return exp_num, source_id


class TestSpectrumListImporter:

    def setup_importer_obj(self, config_helper, input_obj):
        self.sources_labels = ['1D Spectrum at index: 0',
                               '1D Spectrum at index: 1',
                               'Exposure 0, Source ID: 0000',
                               'Exposure 0, Source ID: 1111',
                               'Exposure 1, Source ID: 1111']

        self.sources_data_labels = ['1D Spectrum_index-0',
                                    '1D Spectrum_index-1',
                                    '1D Spectrum_EXP-0_ID-0000',
                                    '1D Spectrum_EXP-0_ID-1111',
                                    '1D Spectrum_EXP-1_ID-1111']

        return FakeSpectrumListImporter(app=config_helper.app,
                                        resolver=config_helper.loaders['object']._obj,
                                        parser=None,
                                        input=input_obj)

    def test_spectrum_list_importer_init_attributes(self, specviz_helper, deconfigged_helper,
                                                    premade_spectrum_list):
        importer_obj = self.setup_importer_obj(specviz_helper, premade_spectrum_list[0])
        assert importer_obj.data_label_default == '1D Spectrum'

        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)
        assert importer_obj.data_label_default == '1D Spectrum'
        assert hasattr(importer_obj, 'sources_items')
        assert hasattr(importer_obj, 'sources_selected')
        assert hasattr(importer_obj, 'sources_multiselect')
        assert hasattr(importer_obj, 'disable_dropdown')
        assert hasattr(importer_obj, 'sources')

        assert importer_obj.sources.selected == [importer_obj.sources.choices[0]]

    # Parameterize to test both single and multiple selection
    @pytest.mark.parametrize('to_select', [['1D Spectrum at index: 1'],
                                           ['1D Spectrum at index: 0',
                                            '1D Spectrum at index: 1',
                                            'Exposure 0, Source ID: 0000',
                                            'Exposure 1, Source ID: 1111']])
    def test_spectrum_list_importer_init_select(self, deconfigged_helper, premade_spectrum_list,
                                                spectrum1d, to_select):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)
        importer_obj.sources.selected = to_select

        assert hasattr(importer_obj.sources, 'items')
        assert hasattr(importer_obj.sources, 'selected')
        assert hasattr(importer_obj.sources, 'multiselect')
        assert hasattr(importer_obj.sources, 'manual_options')

        for index, (spec_dict, spec) in enumerate(
                zip(importer_obj.sources.manual_options, premade_spectrum_list)):

            file_index = str(index)

            # ver, name are stand-ins for exposure and source_id
            # ver == exposure, name == source_id
            if len(spec.meta):
                ver, name = extract_wfss_info(spec)
            else:
                ver, name = str(index), str(index)

            # Apply spectral mask for the data_hash check
            spec = importer_obj._apply_spectral_mask(spec)
            spec_keys = {
                'label': [f"Exposure {ver}, Source ID: {name}",
                          f"1D Spectrum at index: {file_index}"],
                'name': [name, file_index],
                'ver': [ver, file_index],
                'name_ver': [f"{ver}_{name}", file_index],
                'index': [index, index],
                'suffix': [f"EXP-{ver}_ID-{name}", f"index-{file_index}"],
                'data_hash': [create_data_hash(spec), create_data_hash(spec)],
                'obj': None}

            assert isinstance(spec_dict, dict)
            assert len(spec_keys) == len(spec_dict)
            for key in spec_keys:
                assert key in spec_dict
                if key != 'obj':
                    assert spec_dict[key] in spec_keys[key]

            assert isinstance(spec_dict['obj'], Spectrum)
            mask = premade_spectrum_list[index].spectral_axis.mask
            assert np.all(spec_dict['obj'].flux ==
                          premade_spectrum_list[index].flux[~mask])
            assert np.all(spec_dict['obj'].spectral_axis ==
                          premade_spectrum_list[index].spectral_axis[~mask])

    # Method tests
    def test_is_valid(self, deconfigged_helper, premade_spectrum_list,
                      spectrum1d, spectrum2d,
                      make_empty_spectrum, spectrum_collection):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)
        assert importer_obj.is_valid

        importer_obj.input = make_empty_spectrum
        assert not importer_obj.is_valid

        importer_obj.input = spectrum1d
        assert not importer_obj.is_valid

        importer_obj.input = spectrum2d
        assert importer_obj.is_valid

        importer_obj.input = spectrum_collection
        assert importer_obj.is_valid

    def test_on_sources_selected(self, deconfigged_helper, premade_spectrum_list):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)
        # Baseline, default first source selected
        assert importer_obj.sources.selected == [importer_obj.sources.choices[0]]
        assert importer_obj.import_disabled is False

        # Selecting the same thing shouldn't turn off the default flag
        importer_obj.sources.selected = [importer_obj.sources.choices[0]]
        assert importer_obj.import_disabled is False

        importer_obj.sources.selected = []
        assert importer_obj.import_disabled is True

    def test_on_format_selected(self, deconfigged_helper, premade_spectrum_list):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)

        # Baseline, default first source selected
        assert importer_obj.import_disabled is False
        assert importer_obj.sources.selected == [importer_obj.sources.choices[0]]
        importer_obj._on_format_selected_change(change={'new': '1D Spectrum List'})
        # No change
        assert importer_obj.import_disabled is False

        # Set new selection to empty to ensure it doesn't change on format change
        importer_obj.sources.selected = []
        importer_obj._on_format_selected_change(change={'new': '1D Spectrum List'})
        # No selection
        assert importer_obj.import_disabled is True

        # Still no selection
        importer_obj._on_format_selected_change(change={'new': '1D Spectrum Concatenated'})
        assert importer_obj.import_disabled is True

        importer_obj.sources.selected = importer_obj.sources.choices[:2]
        importer_obj._on_format_selected_change(change={'new': '1D Spectrum List'})
        assert importer_obj.import_disabled is False

        # No new selection
        importer_obj._on_format_selected_change(change={'new': '1D Spectrum Concatenated'})
        assert importer_obj.import_disabled is False

        importer_obj._on_format_selected_change(change={'new': 'Not a 1D Spectrum List'})
        assert importer_obj.import_disabled is False

    def test_on_format_selected_2d(self, deconfigged_helper, spectrum2d):
        importer_obj = self.setup_importer_obj(deconfigged_helper, spectrum2d)

        # Baseline, single source selected
        assert importer_obj.import_disabled is False
        importer_obj._on_format_selected_change(change={'new': '1D Spectrum List'})
        # No change
        assert importer_obj.import_disabled is False

        # Reset to no selection
        importer_obj.sources.selected = []
        assert importer_obj.import_disabled is True

        # For 2d spectra, changing format does not change selection, but
        # we specifically enable the import button for 1D Spectrum Concatenated
        importer_obj._on_format_selected_change(change={'new': '1D Spectrum Concatenated'})
        assert importer_obj.import_disabled is False

        importer_obj._on_format_selected_change(change={'new': 'Not a 1D Spectrum List'})
        assert importer_obj.import_disabled is False

    def test_input_to_list_of_spec(self, deconfigged_helper, premade_spectrum_list):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)

        # Test the whole list
        list_results = list(importer_obj.input_to_list_of_spec(premade_spectrum_list))
        assert len(list_results) == len(premade_spectrum_list)
        assert all(isinstance(spec, Spectrum) for spec in list_results)

        # Test individual sources
        for i, spec in enumerate(premade_spectrum_list):
            results = importer_obj.input_to_list_of_spec(spec)
            assert isinstance(results, list)

            result_spec = results[0]
            assert isinstance(result_spec, Spectrum)

            spectral_axis = spec.spectral_axis
            mask = spectral_axis.mask
            # Check both the individual spectra and the results from performing
            # the operation on the whole list
            for result in (result_spec, list_results[i]):
                assert np.all(result.flux == spec.flux[~mask])
                assert np.all(result.spectral_axis == spectral_axis[~mask])
                assert np.all(result.uncertainty.array == spec.uncertainty[~mask].array)
                assert np.all(result.mask == mask[~mask])

    def test_input_to_list_of_spec_not_supported(self, deconfigged_helper, premade_spectrum_list):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)
        with pytest.raises(NotImplementedError, match='not_a_spectrum is not supported'):
            importer_obj.input_to_list_of_spec('not_a_spectrum')

    def test_is_wfssmulti(self, deconfigged_helper, premade_spectrum_list, wfss_spectrum1d):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)
        assert importer_obj.is_wfssmulti(wfss_spectrum1d)

    def test_extract_exposure_sourceid(self, deconfigged_helper, premade_spectrum_list,
                                       wfss_spectrum1d, partially_masked_wfss_spectrum1d,
                                       partially_masked_wfss_spectrum1d_exp1):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)
        exposure, source_id = importer_obj._extract_exposure_sourceid(wfss_spectrum1d)
        assert exposure == '0'
        assert source_id == '0000'

        exposure, source_id = importer_obj._extract_exposure_sourceid(
            partially_masked_wfss_spectrum1d)
        assert exposure == '0'
        assert source_id == '1111'

        exposure, source_id = importer_obj._extract_exposure_sourceid(
            partially_masked_wfss_spectrum1d_exp1)
        assert exposure == '1'
        assert source_id == '1111'

    def test_has_mask(self, deconfigged_helper, premade_spectrum_list, make_empty_spectrum):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)

        # Even unmasked sources still *have* a mask, it's just all False
        for spec in premade_spectrum_list:
            assert importer_obj._has_mask(spec) is True

        assert importer_obj._has_mask(make_empty_spectrum) is False

        make_empty_spectrum.mask = None
        assert importer_obj._has_mask(make_empty_spectrum) is False

    def test_is_fully_masked(self, deconfigged_helper, premade_spectrum_list, spectrum1d):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)

        assert importer_obj._is_fully_masked(spectrum1d) is False
        spectrum1d.mask[:] = True
        assert importer_obj._is_fully_masked(spectrum1d) is True

    def test_is_2d_spectrum(self, deconfigged_helper, spectrum1d, spectrum2d):
        importer_obj = self.setup_importer_obj(deconfigged_helper, spectrum2d)
        assert importer_obj._is_2d_spectrum is True

        importer_obj = self.setup_importer_obj(deconfigged_helper, spectrum1d)
        assert importer_obj._is_2d_spectrum is False

    def test_apply_spectral_mask(self, deconfigged_helper, premade_spectrum_list,
                                 make_empty_spectrum, spectrum1d,
                                 partially_masked_wfss_spectrum1d):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)
        # This does not have a spectral axis mask so it should return as is
        result = importer_obj._apply_spectral_mask(make_empty_spectrum)
        assert result is make_empty_spectrum

        # This doesn't necessarily test something in the spectrum_list code,
        # but it's an error that we should be aware of.
        with pytest.raises(ValueError,
                           match='Spectral axis must be strictly increasing or decreasing.'):
            spectrum1d.spectral_axis.mask[:] = True
            _ = importer_obj._apply_spectral_mask(spectrum1d)

        spec = partially_masked_wfss_spectrum1d
        mask = spec.spectral_axis.mask
        result = importer_obj._apply_spectral_mask(spec)
        assert np.all(result.flux == spec.flux[~mask])
        assert np.all(result.spectral_axis == spec.spectral_axis[~mask])
        assert np.all(result.uncertainty.array == spec.uncertainty[~mask].array)
        assert np.all(result.mask == mask[~mask])

    def test_apply_kwargs(self, deconfigged_helper, premade_spectrum_list):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)

        warning_msg = (
            f"The default source selection ({importer_obj.sources.selected}) will be used.\n"
            f"To load additional sources, please specify them via dropdown or "
            f"as follows:\n'{importer_obj.config}.load(filename, sources = [...]).")

        # Mock the broadcast method to catch the snackbar messages
        with patch.object(deconfigged_helper.app.hub, 'broadcast') as mock_broadcast:
            with pytest.warns(UserWarning, match=re.escape(warning_msg)):
                importer_obj._apply_kwargs({})

            broadcast_msgs = [arg[0][0].text for arg in mock_broadcast.call_args_list
                              if hasattr(arg[0][0], 'text')]
            assert warning_msg in broadcast_msgs

        # No warning this time
        with patch.object(deconfigged_helper.app.hub, 'broadcast') as mock_broadcast:
            importer_obj._apply_kwargs({'sources': ['1D Spectrum at index: 1']})
            assert mock_broadcast.call_count == 0

    def test_output(self, deconfigged_helper, premade_spectrum_list, spectrum1d):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)
        # Must make a selection for output to work
        importer_obj.sources.selected = '1D Spectrum at index: 0'
        assert isinstance(importer_obj.output, list)

        # Check that these are indeed the same
        spec_dict = importer_obj.sources.manual_options[0]
        assert importer_obj.output[0] == spec_dict['obj']

        # Check that is_valid is enforced in the output
        importer_obj.input = spectrum1d
        assert importer_obj.output is None

    @pytest.mark.parametrize('selection', [[],
                                           '1D Spectrum at index: *',
                                           ['1D Spectrum at index: 0',
                                            '1D Spectrum at index: 1',
                                            'Exposure 0, Source ID: 1111',
                                            'Exposure 1, Source ID: 1111']])
    def test_call_method_basic(self, deconfigged_helper, premade_spectrum_list, selection):
        importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)
        sources_data_labels = self.sources_data_labels

        if not selection:
            importer_obj.user_api.sources = selection
            error_msg = "No sources selected."
            # Checking with no selection yet, raises error
            with pytest.raises(
                    ValueError,
                    match=re.escape(error_msg)):
                importer_obj.__call__()

            # Load all anyway
            importer_obj.sources.select_all()
            importer_obj.__call__()
            dc_len = 5
        else:
            importer_obj.user_api.sources = selection
            importer_obj.__call__()
            if isinstance(selection, str):
                dc_len = 2
                sources_data_labels = sources_data_labels[:2]
            elif isinstance(selection, list):
                dc_len = len(selection)
                sources_data_labels = sources_data_labels[:2] + sources_data_labels[-2:]

        # Data collection items
        dc = deconfigged_helper.app.data_collection
        assert len(dc) == dc_len  # number of sources loaded

        assert all([label in sources_data_labels for label in dc.labels])

        # Viewer items
        viewers = deconfigged_helper.viewers
        assert len(viewers) == 1
        assert '1D Spectrum' in viewers

        viewer_dm = viewers['1D Spectrum'].data_menu

        # Note: in a previous version of this test, sources object had a duplicate label hence->
        # TODO: should these be in sync with data collection?
        #  If there is a duplicate data label, it gets overwritten in the viewer
        #  but the data collection will have both.
        assert len(viewer_dm.data_labels_loaded) == len(sources_data_labels)
        assert all([label in sources_data_labels for label in viewer_dm.data_labels_loaded])
        assert len(viewer_dm.data_labels_visible) == len(sources_data_labels)
        assert all([label in sources_data_labels for label in viewer_dm.data_labels_visible])


@pytest.mark.parametrize('with_uncertainty', [True, False])
def test_combine_lists_to_1d_spectrum(with_uncertainty):
    wl = [1, 2, 3] * u.nm
    fnu = [10, 20, 30] * u.Jy
    if with_uncertainty:
        dfnu = [4, 5, 6] * u.Jy
    else:
        dfnu = None

    spec = combine_lists_to_1d_spectrum(wl, fnu, dfnu, u.nm, u.Jy)
    assert isinstance(spec, Spectrum)
    assert isinstance(spec.flux, u.Quantity)
    assert spec.flux.unit == u.Jy

    assert isinstance(spec.spectral_axis, u.Quantity)
    assert spec.spectral_axis.unit == u.nm

    assert np.all(spec.flux.value == np.array([10, 20, 30]))
    assert np.all(spec.spectral_axis.value == np.array([1, 2, 3]))

    if with_uncertainty:
        assert isinstance(spec.uncertainty, StdDevUncertainty)
        assert np.all(spec.uncertainty.array == np.array([4, 5, 6]))
    else:
        assert spec.uncertainty is None


class TestSpectrumListConcatenatedImporter:

    @staticmethod
    def setup_importer_obj(config_helper, input_obj):
        return FakeSpectrumListConcatenatedImporter(app=config_helper.app,
                                                    resolver=config_helper.loaders['object']._obj,
                                                    parser=None,
                                                    input=input_obj)

    def setup_combined_spectrum(self, with_uncertainty):
        wl = [1, 2, 3] * u.nm
        fnu = [10, 20, 30] * u.Jy
        if with_uncertainty:
            dfnu = [4, 5, 6] * u.Jy
        else:
            dfnu = None
        return combine_lists_to_1d_spectrum(wl, fnu, dfnu, u.nm, u.Jy)

    @pytest.mark.parametrize('use_list', [True, False])
    def test_spectrum_list_concatenated_importer_init(self, deconfigged_helper,
                                                      spectrum2d, premade_spectrum_list,
                                                      use_list):
        if use_list:
            importer_obj = self.setup_importer_obj(deconfigged_helper, premade_spectrum_list)
            # Without this, the test gets stuck on the call to importer_obj.output
            # I have not been able to replicate the behavior outside the tests,
            # so I am leaving this test as-is for now.
            importer_obj.sources.selected = []
            assert importer_obj.import_disabled is True
        else:
            importer_obj = self.setup_importer_obj(deconfigged_helper, spectrum2d)
            # Just in case the order is different
            assert len(set(importer_obj.user_api.sources.selected).difference(set(importer_obj.sources.choices))) == 0 # noqa
            assert importer_obj.import_disabled is False

        assert isinstance(importer_obj, SpectrumListImporter)
        # Sneaky negation of boolean to test the disable_dropdown attribute
        assert importer_obj.disable_dropdown is not use_list

        # ensure native units on surface brightness or spectral flux density are converted to flux
        if use_list:
            assert np.all([sp.flux.unit.physical_type == 'spectral flux density'
                           for sp in premade_spectrum_list])
            assert np.all([sp.flux.unit.physical_type == 'spectral flux density'
                           for sp in importer_obj.output])
        else:
            assert spectrum2d.flux.unit.physical_type == 'spectral flux density'
            assert u.Unit(importer_obj.output.flux.unit.to_string()).physical_type == 'spectral flux density' # noqa

    @pytest.mark.parametrize('with_uncertainty', [True, False])
    def test_spectrum_list_concatenated_importer_output(self, deconfigged_helper, with_uncertainty):
        spec = self.setup_combined_spectrum(with_uncertainty)

        importer_obj = self.setup_importer_obj(deconfigged_helper, SpectrumList([spec]))
        result = importer_obj.output

        assert np.all(result.flux == spec.flux)
        assert np.all(result.spectral_axis == spec.spectral_axis)
        if with_uncertainty:
            assert np.all(result.uncertainty.array == spec.uncertainty.array)

    def test_spectrum_list_concatenated_importer_output_empty(self, deconfigged_helper):
        spec = self.setup_combined_spectrum(with_uncertainty=True)

        importer_obj = self.setup_importer_obj(deconfigged_helper, SpectrumList([spec]))
        importer_obj.sources.selected = []
        assert len(importer_obj.output) == 0

    def test_spectrum_list_concatenated_importer_output_default(self, deconfigged_helper):
        spec = self.setup_combined_spectrum(with_uncertainty=True)

        importer_obj = self.setup_importer_obj(deconfigged_helper, SpectrumList([spec]))
        result = importer_obj.output
        assert np.all(result.flux == spec.flux)
        assert np.all(result.spectral_axis == spec.spectral_axis)
        assert np.all(result.uncertainty.array == spec.uncertainty.array)

    def test_spectrum_list_concatenated_importer_output_2d(self, deconfigged_helper, spectrum2d):
        importer_obj = self.setup_importer_obj(deconfigged_helper, spectrum2d)
        _ = importer_obj.output
        assert importer_obj.disable_dropdown is True
        for selected in importer_obj.sources.selected:
            assert isinstance(selected, str)
            assert selected in importer_obj.sources.choices

    def test_spectrum_list_concatenated_importer_call(self, deconfigged_helper):
        spec = self.setup_combined_spectrum(with_uncertainty=True)

        importer_obj = self.setup_importer_obj(deconfigged_helper, SpectrumList([spec]))
        importer_obj.__call__()

        dc = deconfigged_helper.app.data_collection
        assert len(dc) == 1  # 1 concatenated spectrum loaded
        result = dc[0].get_object()

        assert np.all(result.flux == spec.flux)
        assert np.all(result.spectral_axis == spec.spectral_axis)
        assert np.all(result.uncertainty.array == spec.uncertainty.array)



"""
Pytest memlog plugin - log per-test memory usage.

This module provides a pytest plugin for tracking memory usage during test
execution. It supports pytest-xdist for distributed testing and provides
various sorting and filtering options for the memory report.

Memory metrics tracked:
- USS (Unique Set Size): Private memory used by the process
- RSS (Resident Set Size): All physical memory used by the process
- Swap: Swap memory used by the system during test execution

The primary focus is on combined USS + Swap as the key metric for memory
allocation during tests.

Usage
-----
Run pytest with the --memlog option to enable memory logging:

    pytest --memlog                         # Show default 20 tests by USS+Swap diff
    pytest --memlog 10                      # Show top 10 tests by USS+Swap diff
    pytest --memlog 10 --memlog-sort peak   # Sort by peak USS+Swap memory
    pytest --memlog 10 --memlog-max-worker  # Show worker with highest peak memory
"""
import re

import numpy as np
import psutil
import pytest


# ============================================================================
# Module-level storage and dtype
# ============================================================================
_memlog_dtype = np.dtype([('worker_id', 'U20'),
                          ('nodeid', 'U256'),
                          ('uss_before', 'u8'),
                          ('uss_after', 'u8'),
                          ('uss_diff', 'i8'),
                          ('rss_before', 'u8'),
                          ('rss_after', 'u8'),
                          ('rss_diff', 'i8'),
                          ('swap_before', 'u8'),
                          ('swap_after', 'u8'),
                          ('swap_diff', 'i8')])

_full_header = (f'{"uss_before":>12}  {"uss_after":>12}  {"uss_diff":>10}  '
                f'{"rss_before":>12}  {"rss_after":>12}  {"rss_diff":>10}  '
                f'{"swap_before":>12}  {"swap_after":>12}  {"swap_diff":>10}  ')
_after_header = (f'{"uss_after":>12}  '
                 f'{"rss_after":>12}  '
                 f'{"swap_after":>12}  ')
_diff_header = f'{"uss_diff":>10}  {"rss_diff":>10}  {"swap_diff":>10}  '
_full_header_no_worker = _full_header + 'test'
_full_header_with_worker = _full_header + f'{"worker":>8}  test'
_after_header_no_worker = _after_header + 'test'
_after_header_with_worker = _after_header + f'{"worker":>8}  test'

_memlog_records = np.array([], dtype=_memlog_dtype)
_memlog_enabled_flag = False
_default_top_n = 20


# ============================================================================
# Helper functions
# ============================================================================
def _get_memory_bytes():
    """
    Get current process memory usage.

    Returns
    -------
    numpy.void
        A structured array record with rss, swap, and uss fields in bytes.
    """
    p = psutil.Process()
    rss = p.memory_info().rss
    mf = p.memory_full_info()
    swap = getattr(mf, 'swap', 0) or 0
    uss = p.memory_full_info().uss

    mem_array = np.array([(rss, swap, uss)],
                         dtype=[('rss', 'u8'), ('swap', 'u8'), ('uss', 'u8')])
    return mem_array[0]


def _format_bytes(b):
    """
    Format byte count as human-readable string with MiB unit.
    """
    mib = b / (1024.0 * 1024.0)
    return f'{mib:7.2f} MiB'


def _format_memlog_line(record, header='full', include_worker=False):
    """
    Format a memlog record as a display line matching header templates.

    Parameters
    ----------
    record : numpy.void
        A structured array record with memory values and worker_id.
    full_header : bool, optional
        If True (default), display all before/after/diff values.
        If False, display only diff values (uss_diff, rss_diff, swap_diff).
    include_worker : bool, optional
        If True, include the worker ID in the output. Default is False.
    after_only : bool, optional
        If True, display only after values. Default is False.

    Returns
    -------
    str
        Formatted line for display showing memory values aligned with headers.
    """
    # Extract values from record
    uss_before = int(record['uss_before'])
    uss_after = int(record['uss_after'])
    uss_diff = int(record['uss_diff'])
    rss_before = int(record['rss_before'])
    rss_after = int(record['rss_after'])
    rss_diff = int(record['rss_diff'])
    swap_before = int(record['swap_before'])
    swap_after = int(record['swap_after'])
    swap_diff = int(record['swap_diff'])

    # Format all columns matching _full_header layout
    result_str = (f'{_format_bytes(uss_before):>10}  ',
                  f'{_format_bytes(uss_after):>10}  ',
                  f'{_format_bytes(uss_diff):>10}  ',
                  f'{_format_bytes(rss_before):>10}  ',
                  f'{_format_bytes(rss_after):>10}  ',
                  f'{_format_bytes(rss_diff):>10}  ',
                  f'{_format_bytes(swap_before):>10}  ',
                  f'{_format_bytes(swap_after):>10}  ',
                  f'{_format_bytes(swap_diff):>10}  ')

    if header == 'full':
        result_str = ''.join(result_str)
    elif header == 'after':
        # After only
        result_str = result_str[1] + result_str[4] + result_str[7]
    elif header == 'diff':
        # Diffs only
        result_str = result_str[2] + result_str[5] + result_str[8]

    if include_worker:
        worker = record['worker_id']
        return result_str + f'{worker:>8}  {record["nodeid"]}'
    else:
        return result_str + f'{record["nodeid"]}'


def _parse_worker_id(worker_id):
    """
    Parse worker_id into a sortable tuple.

    For xdist worker IDs like 'gw0', 'gw1', etc., returns
    (prefix, number). For 'master', returns ('~', 0) to sort
    last. Ensures numerical ordering within each prefix.

    Parameters
    ----------
    worker_id : str
        The worker ID string to parse.

    Returns
    -------
    tuple
        A (prefix, number) tuple suitable for sorting.
    """
    if worker_id == 'master':
        return '~', 0
    match = re.match(r'([a-z]+)(\d+)', worker_id)
    if match:
        prefix, number = match.groups()
        return prefix, int(number)
    return worker_id, 0


def _extract_memlog_properties(props):
    """
    Extract memory properties from user_properties list into a dict.

    Parameters
    ----------
    props : list
        List of (name, value) tuples from report.user_properties.

    Returns
    -------
    dict
        Dictionary mapping property names to integer values (or string for
        worker_id). Returns empty dict if no memory data found.
    """
    values = {'worker_id': 'master',
              'uss_before': 0,
              'uss_after': 0,
              'uss_diff': 0,
              'rss_before': 0,
              'rss_after': 0,
              'rss_diff': 0,
              'swap_before': 0,
              'swap_after': 0,
              'swap_diff': 0}

    for name, value in props:
        if name in values:
            if name == 'worker_id':
                values[name] = value
            else:
                values[name] = int(value)

    # Check if we have any memory data
    has_memory_data = any(
        values[k] != 0 for k in ['uss_before', 'uss_after', 'uss_diff',
                                 'rss_before', 'rss_after', 'rss_diff',
                                 'swap_before', 'swap_after', 'swap_diff'])

    return values if has_memory_data else {}


def _apply_memlog_sort(records, sort_method, top_n):
    """
    Apply sorting to memlog records based on sort_method.

    All methods that reference USS implicitly include the combined USS + Swap
    metric as the primary indicator of memory allocation.

    Parameters
    ----------
    records : numpy.ndarray
        Array of memlog records.
    sort_method : str
        Sorting method:
        - 'diff': Sort by USS + Swap difference (memory allocated during test)
        - 'before': Sort by highest USS before test
        - 'after': Sort by highest USS after test
        - 'peak': Sort by highest peak combined USS + Swap
        - 'seq': Reverse chronological order (test execution order)
    top_n : int
        Number of top records to return.

    Returns
    -------
    numpy.ndarray
        Sorted records, limited to top_n items.
    """
    if len(records) == 0:
        return records

    if sort_method == 'diff':
        combined_diffs = records['uss_diff'] + records['swap_diff']
        sort_idx = np.argsort(-combined_diffs)

    elif sort_method == 'before':
        sort_idx = np.argsort(-records['uss_before'])

    elif sort_method == 'after':
        sort_idx = np.argsort(-records['uss_after'])

    elif sort_method == 'peak':
        uss_peaks = np.maximum(records['uss_before'], records['uss_after'])
        swap_peaks = np.maximum(records['swap_before'], records['swap_after'])
        combined_peaks = uss_peaks + swap_peaks
        sort_idx = np.argsort(-combined_peaks)

    elif sort_method == 'seq':
        sort_idx = np.arange(len(records) - 1, -1, -1)

    else:
        sort_idx = np.arange(len(records))

    sorted_records = records[sort_idx]
    return sorted_records[:top_n]


# ============================================================================
# Pytest hooks
# ============================================================================

def memlog_addoption(parser):
    """
    Register memlog command-line options.
    """
    group = parser.getgroup('memlog')
    group.addoption(
        '--memlog',
        action='store',
        nargs='?',
        const=str(_default_top_n),
        dest='memlog',
        default='',
        help='Enable per-test memory logging and summary.\n'
             'Usage: --memlog [N], where N is the number of top entries to display.\n'
             f'Default: {_default_top_n}')

    group.addoption(
        '--memlog-sort',
        action='store',
        dest='memlog_sort',
        default='diff',
        help=('Sorting method for memory (USS + Swap) results. Default: diff\n'
              'diff   - Sort by USS + Swap difference (memory allocated during test).\n'
              'before - Sort by highest USS before test.\n'
              'after  - Sort by highest USS after test.\n'
              'peak   - Sort by highest peak combined USS + Swap.\n'
              'seq    - Sort by test execution order (helps identify sustained memory).\n'
              'worker - Group by worker ID (xdist only), then sort by peak USS + Swap.'))

    group.addoption(
        '--memlog-max-worker',
        action='store_true',
        dest='memlog_max_worker',
        default=False,
        help=('Show memory report for the worker with highest peak '
              'memory allocation (xdist only).'))


def memlog_configure(config):
    """
    Configure memlog based on command-line options.

    This function initializes the memlog plugin settings. It should be
    called from conftest.py's pytest_configure hook.

    Parameters
    ----------
    config : pytest.Config
        The pytest configuration object.
    """
    global _memlog_enabled_flag

    if config.getoption('memlog') or config.getoption('memlog_max_worker'):
        config._memlog_top = config.getoption('memlog')
        config._memlog_sort = config.getoption('memlog_sort')
        config._memlog_max_worker = config.getoption('memlog_max_worker')
        _memlog_enabled_flag = True
        config._memlog_enabled = True


def memlog_runtest_setup(item):
    """
    Setup hook that records memory before test.
    """
    if not _memlog_enabled_flag:
        return
    item._mem_before = _get_memory_bytes()


def memlog_runtest_teardown(item, _):
    """
    Teardown hook that records memory after test.
    """
    if not _memlog_enabled_flag:
        return
    item._mem_after = _get_memory_bytes()


@pytest.hookimpl(hookwrapper=True)
def memlog_runtest_makereport(item, call):
    """
    Hook wrapper to attach memory measurements to report user_properties.

    This runs during report creation when we still have access to the item.
    The user_properties are serialized and sent to master in xdist. We track
    USS, RSS, and Swap for each test, though the analysis focuses on
    USS + Swap as the primary memory allocation metric.
    """
    outcome = yield
    report = outcome.get_result()

    if call.when != 'teardown':
        return

    if not _memlog_enabled_flag:
        return

    mem_before = getattr(item, '_mem_before', None)
    mem_after = getattr(item, '_mem_after', None)

    if mem_before is None or mem_after is None:
        return

    # Attach to user_properties - these get serialized to master in xdist
    for prefix in ('uss', 'rss', 'swap'):
        before = int(mem_before[prefix])
        after = int(mem_after[prefix])
        diff = after - before

        report.user_properties.append((f'{prefix}_before', before))
        report.user_properties.append((f'{prefix}_after', after))
        report.user_properties.append((f'{prefix}_diff', diff))

    # Get worker_id from config (xdist sets this)
    worker_id = getattr(item.config, 'workerinput', {}).get('workerid', 'master')
    report.user_properties.append(('worker_id', worker_id))


def memlog_runtest_logreport(report):
    """
    Log report hook that collects memory measurements from user_properties.

    This runs on both workers and master. On master (xdist), it receives
    the serialized user_properties from workers.
    """
    global _memlog_records

    if report.when != 'teardown':
        return

    props = getattr(report, 'user_properties', [])

    if not props:
        return

    mem_props = _extract_memlog_properties(props)
    if not mem_props:
        return

    # Build record tuple in dtype field order
    record_values = (mem_props['worker_id'],
                     getattr(report, 'nodeid', '<unknown>'),
                     mem_props['uss_before'],
                     mem_props['uss_after'],
                     mem_props['uss_diff'],
                     mem_props['rss_before'],
                     mem_props['rss_after'],
                     mem_props['rss_diff'],
                     mem_props['swap_before'],
                     mem_props['swap_after'],
                     mem_props['swap_diff'])

    record = np.array([record_values], dtype=_memlog_dtype)
    _memlog_records = np.append(_memlog_records, record)


def _display_max_worker_report(terminalreporter, records, top_n):
    """
    Display memory report for the worker with the highest peak memory.

    Identifies the worker that had the highest peak combined USS + Swap memory
    allocation across all tests, then displays all that worker's tests in
    execution order.

    Parameters
    ----------
    terminalreporter : TerminalReporter
        The pytest terminal reporter.
    records : numpy.ndarray
        Array of memlog records.
    top_n : int
        Number of top records to display.
    """
    if len(records) == 0:
        terminalreporter.write_line('memlog: no worker found with memory data.')
        return

    # Find the worker with the highest peak USS + Swap memory across all tests
    # using vectorized operations
    uss_peaks = np.maximum(records['uss_before'], records['uss_after'])
    swap_peaks = np.maximum(records['swap_before'], records['swap_after'])
    combined_peaks = uss_peaks + swap_peaks
    max_idx = np.argmax(combined_peaks)
    max_peak = int(combined_peaks[max_idx])
    max_worker = records[max_idx]['worker_id']

    # Filter to only this worker's records
    worker_records = records[records['worker_id'] == max_worker]

    # Sort by execution order (seq) to see memory allocation pattern
    worker_records = _apply_memlog_sort(worker_records, 'seq', top_n)

    title = (f'Top {top_n} tests for worker {max_worker}, in sequence order '
             f'(highest peak USS + Swap: {_format_bytes(max_peak)})')
    terminalreporter.write_sep('-', title)

    terminalreporter.write_line(_full_header_no_worker)

    for r in worker_records:
        terminalreporter.write_line(_format_memlog_line(r))

    # Display peak memory usage across all tests
    _display_peak_usage(terminalreporter, records)

    terminalreporter.write_sep('-', 'end of memlog summary')


def _display_worker_grouped_report(terminalreporter, records, top_n):
    """
    Display memory report grouped by worker ID and sorted by test execution.

    This report shows the top N tests for each worker in the order they were
    executed, making it easier to identify memory leaks or accumulated memory
    allocation over the test sequence for each worker.

    Parameters
    ----------
    terminalreporter : TerminalReporter
        The pytest terminal reporter.
    records : numpy.ndarray
        Array of memlog records.
    top_n : int
        Number of top records to display per worker.
    """
    if len(records) == 0:
        return

    # Get unique worker IDs and sort them numerically
    unique_workers = np.unique(records['worker_id'])
    sorted_workers = sorted(unique_workers, key=_parse_worker_id)

    # Display results grouped by worker in execution order
    title = f'Top {top_n} tests by worker, in test execution order'
    terminalreporter.write_sep('-', title)

    for worker_id in sorted_workers:
        # Filter records for this worker and sort by execution order (seq)
        worker_records = records[records['worker_id'] == worker_id]
        worker_records = _apply_memlog_sort(worker_records, 'seq', top_n)

        terminalreporter.write_line(f'\nWorker: {worker_id}')
        terminalreporter.write_line(_full_header_with_worker)

        for r in worker_records[:top_n]:
            terminalreporter.write_line(_format_memlog_line(r, include_worker=True))


def _display_standard_report(terminalreporter, records, sort_method, top_n):
    """
    Display standard memory report with selected sorting.

    Displays all memory before/after/diff values aligned with the full header
    template. Sorting is based on the USS + Swap combined metric.

    Parameters
    ----------
    terminalreporter : TerminalReporter
        The pytest terminal reporter.
    records : numpy.ndarray
        Array of memlog records.
    sort_method : str
        Sorting method: 'diff', 'before', 'after', 'peak', or 'seq'.
    top_n : int
        Number of top records to display.
    """
    # Apply the selected sort method to all records
    records = _apply_memlog_sort(records, sort_method, top_n)

    # Create descriptive title based on sort method
    sort_titles = {
        'diff': 'by USS + Swap difference',
        'before': 'by USS before test',
        'after': 'by USS after test',
        'peak': 'by peak USS + Swap',
        'seq': 'by test execution order'
    }
    sort_desc = sort_titles.get(sort_method, 'by memory difference')
    title = f'Top {top_n} tests {sort_desc}'
    terminalreporter.write_sep('-', title)

    terminalreporter.write_line(_diff_header)

    for r in records:
        terminalreporter.write_line(_format_memlog_line(r, header='diff', include_worker=True))


def _calculate_peak_usage(records):
    """
    Calculate peak memory usage across all records using vectorized operations.

    Parameters
    ----------
    records : numpy.ndarray
        Array of memlog records.

    Returns
    -------
    dict
        Dictionary with 'uss', 'rss', 'swap' peak values.
    """
    if len(records) == 0:
        return {'uss': 0, 'rss': 0, 'swap': 0}

    # Use vectorized operations to find max peaks for each memory type
    peaks = {}
    for prefix in ('uss', 'rss', 'swap'):
        before = records[f'{prefix}_before']
        after = records[f'{prefix}_after']
        peaks[prefix] = int(np.maximum(before, after).max())

    return peaks


def _display_peak_usage(terminalreporter, records):
    """
    Display peak memory usage summary across all workers.

    Identifies and displays the test(s) that reached the highest combined USS +
    Swap memory allocation (peak value) in each worker. Shows the test with
    maximum memory usage for each worker separately.

    Parameters
    ----------
    terminalreporter : TerminalReporter
        The pytest terminal reporter.
    records : numpy.ndarray
        Array of memlog records.
    """
    if len(records) == 0:
        return

    terminalreporter.write_line('')
    terminalreporter.write_line('Peak memory usage (maximum USS + Swap per worker):')

    # Get unique worker IDs and sort them numerically
    unique_workers = np.unique(records['worker_id'])
    sorted_workers = sorted(unique_workers, key=_parse_worker_id)

    terminalreporter.write_line(_full_header_with_worker)

    # For each worker, find the test with the maximum combined peak
    for worker_id in sorted_workers:
        worker_records = records[records['worker_id'] == worker_id]

        # Calculate peak values for this worker's tests
        uss_swap_before = worker_records['uss_before'] + worker_records['swap_before']
        uss_swap_after = worker_records['uss_after'] + worker_records['swap_after']
        combined_peaks = np.maximum(uss_swap_before, uss_swap_after)

        # Find the test with the maximum combined peak in this worker
        max_idx = np.argmax(combined_peaks)
        peak_record = worker_records[max_idx]

        terminalreporter.write_line(_format_memlog_line(peak_record,
                                                        include_worker=True))


def _display_final_usage(terminalreporter, records):
    """
    Display final memory usage summary of final tests in all workers.

    Shows the memory state (after the test, before teardown) of the last test
    executed in each worker, helping identify final memory footprint across
    distributed workers.

    Parameters
    ----------
    terminalreporter : TerminalReporter
        The pytest terminal reporter.
    records : numpy.ndarray
        Array of memlog records.
    """
    if len(records) == 0:
        return

    terminalreporter.write_line('')
    terminalreporter.write_line('Final memory usage '
                                '(last test in each worker, after test execution):')

    # Get unique worker IDs and sort them numerically
    unique_workers = np.unique(records['worker_id'])
    sorted_workers = sorted(unique_workers, key=_parse_worker_id)

    terminalreporter.write_line(_after_header_with_worker)

    # For each worker, get the last test executed (highest index in execution order)
    for worker_id in sorted_workers:
        worker_records = records[records['worker_id'] == worker_id]
        # The last record for this worker is the one with the highest index
        last_record = worker_records[-1]
        terminalreporter.write_line(_format_memlog_line(last_record,
                                                        header='after',
                                                        include_worker=True))


def memlog_terminal_summary(terminalreporter, config=None):
    """
    Terminal summary hook that prints memlog summary.
    """
    if config is None:
        config = terminalreporter.config

    if not getattr(config, '_memlog_enabled', False):
        return

    if len(_memlog_records) == 0:
        terminalreporter.write_line('memlog: no records collected.')
        return

    top_n = getattr(config, '_memlog_top', '20')
    if isinstance(top_n, str):
        if top_n.isdigit():
            top_n = int(top_n)
        else:
            top_n = 20
    elif not isinstance(top_n, int):
        top_n = 20

    if len(_memlog_records) == 0:
        terminalreporter.write_line('memlog: no records collected.')
        return

    sort_method = getattr(config, '_memlog_sort', 'diff')

    # If max worker is requested, find and report on the worker with
    # the highest peak memory allocation
    if getattr(config, '_memlog_max_worker', False):
        _display_max_worker_report(terminalreporter, _memlog_records, top_n)

    else:
        # Group by worker_id if sorting by worker
        if sort_method == 'worker':
            _display_worker_grouped_report(terminalreporter, _memlog_records, top_n)
        else:
            _display_standard_report(terminalreporter, _memlog_records, sort_method, top_n)

    # Display peak memory usage across all tests
    _display_peak_usage(terminalreporter, _memlog_records)

    # Display final memory usage of last test in each worker
    _display_final_usage(terminalreporter, _memlog_records)

    terminalreporter.write_sep('-', 'end of memlog summary')



import numpy as np
import astropy.units as u
from astropy.modeling.fitting import LevMarLSQFitter
from astropy.modeling.models import Gaussian1D
from astropy.tests.helper import assert_quantity_allclose
import pytest
from specutils import Spectrum


@pytest.mark.parametrize("helper", ('specviz2d_helper', 'deconfigged_helper'))
def test_cross_dispersion_profile(helper, request):
    """Tests the basic functionality of the cross dispersion profile plugin."""
    helper = request.getfixturevalue(helper)

    # Create a 2D array representing a series of Gaussian profiles.
    # Each column i corresponds to a Gaussian centered at y=5 with amplitude=i.
    # The amplitude increases along the spectral axis
    x = np.arange(25)
    y = np.arange(10)
    arr = np.zeros((10, 25))
    for i in range(25):
        arr[:, i] = Gaussian1D(amplitude=i, mean=5, stddev=1)(y)

    data = Spectrum(flux=arr * u.Jy, spectral_axis=x * u.nm)
    helper.load(data, format='2D Spectrum')

    cdp = helper.plugins['Cross Dispersion Profile']
    cdp.keep_active = True

    fitter = LevMarLSQFitter()
    model = Gaussian1D()

    # check defaults
    assert cdp.pixel == 12
    assert cdp.use_full_width
    assert cdp.width == 10
    assert cdp.y_pixel == 5

    # check that 2d spectrum viewer marks marks (vertical line with scatter)
    # are in the correct starting location
    assert np.all(cdp._obj.marks['2d']['pix'].marks_list[0].x == 12)
    assert np.all(cdp._obj.marks['2d']['y_pix'].marks_list[0].y == 5)

    # make sure the profile at row 12 is a gaussian with amp. 12, as the test
    # data was designed
    fit = fitter(model, y, cdp.profile)
    assert fit.amplitude == 12 * u.Jy
    assert fit.mean == 5

    # now move the pixel slider to col 13 and make sure the fit profile
    # now reflects a gaussian with amplitude 13, and that the plugin marks
    # have moved as well
    cdp.pixel = 13
    profile_13 = cdp.profile
    fit = fitter(model, y, profile_13)
    assert fit.amplitude == 13 * u.Jy
    assert fit.mean == 5
    assert np.all(cdp._obj.marks['2d']['pix'].marks_list[0].x == 13)
    assert np.all(cdp._obj.marks['2d']['y_pix'].marks_list[0].y == 5)

    # and now adjust the width, and make sure the profile and
    # marks changed accordingly
    cdp.use_full_width = False
    cdp.width = 5
    assert len(cdp.profile) == 5
    assert len(cdp._obj.marks['2d']['pix'].marks_list[0].y) == 5
    cdp.width = 7
    prev_profile = cdp.profile  # save profile to compare after unit conversions
    assert len(cdp.profile) == 7
    assert len(cdp._obj.marks['2d']['pix'].marks_list[0].y) == 7

    # test flux unit conversion
    uc = helper.plugins['Unit Conversion']
    uc.flux_unit.selected = 'MJy'
    assert cdp.profile.unit == u.MJy
    # convert profile saved at the step above to new
    # unit to make sure it was converted
    assert_quantity_allclose(prev_profile, cdp.profile)

    # test flux unit conversion that needs an equivalency
    uc = helper.plugins['Unit Conversion']
    uc.flux_unit.selected = 'erg / (Angstrom s cm2)'
    un = u.Unit('erg / (Angstrom s cm2)')
    assert cdp.profile.unit == un
    # convert before comparing since we need an equivalency based on the wavelength
    eqv = u.spectral_density(cdp._obj.wav * u.Unit(cdp._obj.sa_display_unit))
    converted = cdp.profile.to(u.Jy, eqv)
    assert_quantity_allclose(prev_profile, converted)

    # do a spectral axis conversion and make sure the plot title updated
    # to reflect the new wavelength
    uc.spectral_unit.selected = 'nm'
    assert 'nm' in cdp._obj.plot.figure.title

    # todo: add tests to cover selecting new datasets (e.g background image
    # from the spectral extraction plugin) once JDAT-5426 is resolved. need
    # to test that dataset_selected behaves correctly



from jdaviz import __version__
from jdaviz.pytest_memlog import (memlog_addoption, memlog_configure, memlog_runtest_setup,
                                  memlog_runtest_teardown, memlog_runtest_makereport,
                                  memlog_runtest_logreport, memlog_terminal_summary)


# ============================================================================
# Memory logging plugin (memlog) - imported from pytest_memlog.py
# ============================================================================
def pytest_addoption(parser):
    """
    Register pytest options.
    """
    memlog_addoption(parser)


def pytest_runtest_setup(item):
    """
    Setup hook that records memory before test.
    """
    memlog_runtest_setup(item)


def pytest_runtest_teardown(item, nextitem):
    """
    Teardown hook that records memory after test.
    """
    memlog_runtest_teardown(item, nextitem)


# Re-export the hookwrapper directly
pytest_runtest_makereport = memlog_runtest_makereport


def pytest_runtest_logreport(report):
    """
    Log report hook that collects memory measurements from user_properties.
    """
    memlog_runtest_logreport(report)


def pytest_terminal_summary(terminalreporter, config=None):
    """
    Terminal summary hook that prints memlog summary.
    """
    memlog_terminal_summary(terminalreporter, config)


try:
    from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
except ImportError:
    PYTEST_HEADER_MODULES = {}
    TESTED_VERSIONS = {}


# This is repeated from jdaviz/conftest.py because tox cannot grab test
# header from that file.
def pytest_configure(config):
    # Initialize memlog
    memlog_configure(config)

    PYTEST_HEADER_MODULES['astropy'] = 'astropy'
    PYTEST_HEADER_MODULES['pyyaml'] = 'yaml'
    PYTEST_HEADER_MODULES['scikit-image'] = 'skimage'
    PYTEST_HEADER_MODULES['specutils'] = 'specutils'
    PYTEST_HEADER_MODULES['specreduce'] = 'specreduce'
    PYTEST_HEADER_MODULES['asteval'] = 'asteval'
    PYTEST_HEADER_MODULES['echo'] = 'echo'
    PYTEST_HEADER_MODULES['idna'] = 'idna'
    PYTEST_HEADER_MODULES['traitlets'] = 'traitlets'
    PYTEST_HEADER_MODULES['bqplot'] = 'bqplot'
    PYTEST_HEADER_MODULES['bqplot-image-gl'] = 'bqplot_image_gl'
    PYTEST_HEADER_MODULES['glue-core'] = 'glue'
    PYTEST_HEADER_MODULES['glue-jupyter'] = 'glue_jupyter'
    PYTEST_HEADER_MODULES['glue-astronomy'] = 'glue_astronomy'
    PYTEST_HEADER_MODULES['ipyvue'] = 'ipyvue'
    PYTEST_HEADER_MODULES['ipyvuetify'] = 'ipyvuetify'
    PYTEST_HEADER_MODULES['ipysplitpanes'] = 'ipysplitpanes'
    PYTEST_HEADER_MODULES['ipygoldenlayout'] = 'ipygoldenlayout'
    PYTEST_HEADER_MODULES['ipypopout'] = 'ipypopout'
    PYTEST_HEADER_MODULES['solara'] = 'solara'
    PYTEST_HEADER_MODULES['vispy'] = 'vispy'
    PYTEST_HEADER_MODULES['gwcs'] = 'gwcs'
    PYTEST_HEADER_MODULES['asdf'] = 'asdf'
    PYTEST_HEADER_MODULES['stdatamodels'] = 'stdatamodels'
    PYTEST_HEADER_MODULES['roman_datamodels'] = 'roman_datamodels'

    TESTED_VERSIONS['jdaviz'] = __version__
